# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""Web Server Gateway Interface configuration

This module exposes the WSGI callable as a module-level variable named
``application``.

For more information on this file, see
https://api_docs.djangoproject.com/en/3.0/howto/deployment/wsgi/
"""

import os

from django.core.wsgi import get_wsgi_application

os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'broker_web.main.settings')

application = get_wsgi_application()



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""Adds command to ``manage.py`` for generating a new secret key"""

import string

from django.core.management.base import BaseCommand
from django.utils.crypto import get_random_string


class Command(BaseCommand):
    help = 'Generates and prints a new secret key'

    @staticmethod
    def gen_secret_key(length=50, chars=None):
        """Generate a secret key for Django

        Args:
            length (int): The length of the key
            chars  (str): Characters to use
        """

        return get_random_string(length, chars)

    def add_arguments(self, parser):
        """Add arguments to a command line parser

        Adds ``len`` and ``chars`` arguments to CLI.

        Args:
            parser: The parser to add arguments to
        """

        parser.add_argument('len', nargs=1, type=int, default=50)

    def handle(self, *args, **options):
        """Handle a command line call for the parent class"""

        length = options['len'][0]
        chars = string.ascii_lowercase + string.digits + string.punctuation
        self.stdout.write(self.gen_secret_key(length, chars))



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""The ``main`` module represents the main django application. It handles the
settings and URL routing for the website and includes minimal custom views or
models.
"""

from . import asgi, settings, urls, views, wsgi



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""A django application for serving up data about recently observed LSST/ZTF
alerts published in a collection of configured PubSub topics.
"""

# Todo alerts:
# - Add / use PubSub configuration to / from global settings
# - Populate the AlertsJsonView view using PubSub messages
# - Use arguments from the filter results from to filter values in AlertsJsonView
# - Move ``forms.topics`` into global settings



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""URL routing for the main django application. URLS from each application
is routed to the following namespaces:

+-------------------------+---------------------+
| Application             | Namespace           |
+=========================+=====================+
|``apps.alerts``          | ``alerts``          |
+-------------------------+---------------------+
|``apps.contact``         | ``contact``         |
+-------------------------+---------------------+
|``apps.getting_started`` | ``getting-started`` |
+-------------------------+---------------------+
|``apps.objects``         | ``objects``         |
+-------------------------+---------------------+
|``apps.signup``          | ``signup``          |
+-------------------------+---------------------+
|``apps.subscriptions``   | ``subscriptions``   |
+-------------------------+---------------------+
"""

from django.contrib import admin
from django.urls import include
from django.urls import path

from .views import IndexView, why_pgb_view

urlpatterns = [
    # URLs for custom views
    path('', IndexView.as_view(), name='home'),
    path('why_pgb/', why_pgb_view, name='why-pgb'),
    path('contact/', include('broker_web.apps.contact.urls', namespace='contact')),
    path('alerts/', include('broker_web.apps.alerts.urls', namespace='alerts')),
    path('objects/', include('broker_web.apps.objects.urls', namespace='objects')),
    path('getting_started/', include('broker_web.apps.getting_started.urls', namespace='getting-started')),
    path('signup/', include('broker_web.apps.signup.urls', namespace='signup')),
    path('subscriptions/', include('broker_web.apps.subscriptions.urls', namespace='subscriptions')),

    # Built in Django URL patterns
    path('users/', include('django.contrib.auth.urls')),
    path('admin/', admin.site.urls)
]

# Error handling
handler404 = 'broker_web.apps.error_pages.views.error_404'
handler500 = 'broker_web.apps.error_pages.views.error_500'



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""The ``forms`` module defines views forms for data entry and query
construction.

.. autosummary::
   :nosignatures:

   broker_web.apps.objects.forms.FilterObjectsForm
"""

from django import forms

topics = (
    'ztf_all',
    '91bg',
    'sne Ia',
    'CV'
)


# Todo: This form is almost identical to the FilterAlertsForm from the
# alerts app. I expect moving forward the two will diverge, but if that
# is not the case, consider making things more dry
class FilterObjectsForm(forms.Form):
    """Form for filtering a table of alerted objects

    Fields:
        time_range (``DurationField``)
        min_ra (``FloatField``)
        max_ra (``FloatField``)
        min_dec (``FloatField``)
        max_dec (``FloatField``)
    """

    time_range = forms.DurationField(required=False, label='Publication time')
    min_ra = forms.FloatField(
        required=False,
        label='Min RA',
        widget=forms.TextInput()
    )

    max_ra = forms.FloatField(required=False, label='Max RA', widget=forms.TextInput())
    min_dec = forms.FloatField(required=False, label='Min Dec', widget=forms.TextInput())
    max_dec = forms.FloatField(required=False, label='Max Dec', widget=forms.TextInput())



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""Tests for the ``urls`` module."""

from django.test import TestCase
from django.urls import resolve, reverse

from broker_web.apps.subscriptions import urls, views


class TestUrlRouting(TestCase):
    """Test URLs are routed to the correct views"""

    app_name = urls.app_name

    def test_subscriptions_routing(self):
        """Test 'subscriptions' is routed to``SubscriptionsView``"""

        url = reverse(f'{self.app_name}:subscriptions')
        self.assertEqual(views.SubscriptionsView, resolve(url).func.view_class)

    def test_profile_routing(self):
        """Test 'profile' is routed to``ProfileView``"""

        url = reverse(f'{self.app_name}:profile')
        self.assertEqual(views.ProfileView, resolve(url).func.view_class)



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""Tests for the ``forms`` module."""

from django.test import TestCase

from broker_web.apps.contact.forms import ContactForm


class TestContactForm(TestCase):
    """Test the ``ContactForm`` validates correctly"""

    def test_valid_form(self):
        """Test a full form is valid invalid"""

        form = ContactForm(data=dict(
            email='test@email.com',
            name='Test Name',
            message='Test message.'
        ))
        self.assertFalse(form.is_valid())

    def test_invalid_email(self):
        """Test email validation is enforced"""

        form = ContactForm(data=dict(
            email='invalid_email',
            name='Test Name',
            message='Test message.'
        ))
        self.assertFalse(form.is_valid())

    def test_empty_form(self):
        """Test an empty form is invalid"""

        form = ContactForm(data=dict(
            email='',
            name='',
            message=''
        ))
        self.assertFalse(form.is_valid())



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""The ``admin`` module defines app level interfaces for the website's
admin panel.
"""

from django.contrib import admin
from django.contrib.auth.admin import UserAdmin

from .forms import CustomUserChangeForm, CustomUserCreationForm
from .models import CustomUser


class CustomUserAdmin(UserAdmin):
    """Custom admin interface for user management"""

    add_form = CustomUserCreationForm
    form = CustomUserChangeForm
    model = CustomUser

    # Fields shown in the main summary page
    list_display = ('email', 'first_name', 'last_name', 'country', 'affiliation')

    # Fields shown when editing a new user
    fieldsets = (
        ('User Data', {'fields': list_display}),
        ('Permissions', {'fields': ('is_staff', 'is_active')}),
    )

    # Fields shown when creating a new user
    add_fieldsets = (
        (None, {'fields': ('email', 'password1', 'password2', 'is_staff', 'is_active')}),
    )

    search_fields = ('email', 'first_name', 'last_name')
    ordering = ('email',)


admin.site.register(CustomUser, CustomUserAdmin)



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""The ``forms`` module defines views forms for data entry and query
construction.

.. autosummary::
   :nosignatures:

   broker_web.apps.signup.forms.CustomUserChangeForm
   broker_web.apps.signup.forms.CustomUserCreationForm
"""

from captcha.fields import ReCaptchaField
from django.contrib.auth.forms import UserChangeForm, UserCreationForm

from .models import CustomUser

user_fields = 'email', 'first_name', 'last_name', 'country', 'affiliation'


class CustomUserCreationForm(UserCreationForm):
    """Custom form for creating a new ``CustomUser``"""

    captcha = ReCaptchaField()

    class Meta(UserCreationForm.Meta):
        model = CustomUser
        fields = user_fields

    def validate(self, value):
        """Check if value consists only of valid emails."""
        print(value)
        super().validate(value)


class CustomUserChangeForm(UserChangeForm):
    """Custom form for modifying user data"""

    class Meta:
        model = CustomUser
        fields = user_fields



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Consumer class to pull Pub/Sub messages via a Python client, and work with data.

Pub/Sub Python Client docs: https://googleapis.dev/python/pubsub/latest/index.html

Used by `BrokerStreamPython`, but can be called independently.

Use-case: Save alerts to a database

    The demo for this implementation assumes that the real use-case is to save alerts
    to a database rather than view them through a TOM site.
    Therefore, the `Consumer` currently saves the alerts in real time,
    and then simply returns a list of alerts after all messages are processed.
    That list is then coerced into an iterator by the `Broker`.
    If the user really cares about the `Broker`'s iterator, `stream_alerts` can
    be tweaked to yield the alerts in real time.

Basic workflow:

.. code:: python

    consumer = ConsumerStreamPython(subscription_name)

    alert_dicts_list = consumer.stream_alerts(
        user_filter=user_filter,
        **kwargs,
    )
    # alerts are processed and saved in real time. the list is returned for convenience.

See especially:

.. autosummary::
   :nosignatures:

   ConsumerStreamPython.authenticate
   ConsumerStreamPython.touch_subscription
   ConsumerStreamPython.stream_alerts
   ConsumerStreamPython.callback
   ConsumerStreamPython.save_alert

"""

# from concurrent.futures.thread import ThreadPoolExecutor
from django.conf import settings
from google.api_core.exceptions import NotFound
from google.cloud import pubsub_v1
# from google.cloud.pubsub_v1.subscriber.scheduler import ThreadScheduler
from google import auth
from google.cloud import logging as gc_logging
from google_auth_oauthlib.helpers import credentials_from_session
from requests_oauthlib import OAuth2Session
import queue
from queue import Empty

from .utils.templatetags.utility_tags import avro_to_dict


PITTGOOGLE_PROJECT_ID = "ardent-cycling-243415"


class ConsumerStreamPython:
    """Consumer class to manage Pub/Sub connections and work with messages.

    Initialization does the following:

        - Authenticate the user via OAuth 2.0.

        - Create a `google.cloud.pubsub_v1.SubscriberClient` object.

        - Create a `queue.Queue` object to communicate with the background thread
          running the streaming pull.

        - Make sure the subscription exists and we can connect. Create it, if needed.

    To view logs, visit: https://console.cloud.google.com/logs

        - Make sure you are logged in, and your project is selected in the dropdown
          at the top.

        - Click the "Log name" dropdown and select the subscription name you
          instantiate this consumer with.

    TODO: Give the user a standard logger.
    """

    def __init__(self, subscription_name, ztf_fields=None):
        """Authenticate user; create client; set subscription path; check connection."""
        user_project = settings.GOOGLE_CLOUD_PROJECT
        self.database_list = []  # list of dicts. fake database for demo.

        # authenticate the user
        self.get_credentials(user_project)

        # instantiate client
        self.client = pubsub_v1.SubscriberClient(credentials=self.credentials)

        # logger
        log_client = gc_logging.Client(
            project=user_project, credentials=self.credentials
        )
        self.logger = log_client.logger(subscription_name)

        # subscription
        self.subscription_name = subscription_name
        self.subscription_path = f"projects/{user_project}/subscriptions/{subscription_name}"
        # Topic to connect the subscription to, if it needs to be created.
        # If the subscription already exists but is connected to a different topic,
        # the user will be notified and this topic_path will be updated for consistency.
        self.topic_path = f"projects/{PITTGOOGLE_PROJECT_ID}/topics/{subscription_name}"
        self.touch_subscription()

        self._set_ztf_fields(ztf_fields)

        self.queue = queue.Queue()  # queue for communication between threads

        # for the TOM `GenericAlert`. this won't be very helpful without instructions.
        self.pull_url = "https://pubsub.googleapis.com/v1/{subscription_path}"

    def get_credentials(self, user_project):
        """Create user credentials object from service account credentials or an OAuth.

        Try service account credentials first. Fall back to OAuth.
        """
        try:
            self.credentials, project = auth.load_credentials_from_file(
                settings.GOOGLE_APPLICATION_CREDENTIALS
            )
            assert project == user_project  # TODO: handle this better

        except auth.exceptions.DefaultCredentialsError:
            self.authenticate_with_oauth()
            self.credentials = credentials_from_session(self.oauth2)

    def authenticate_with_oauth(self):
        """Guide user through authentication; create `OAuth2Session` for credentials.

        The user will need to visit a URL, authenticate themselves, and authorize
        `PittGoogleConsumer` to make API calls on their behalf.

        The user must have a Google account that is authorized make API calls
        through the project defined by the `GOOGLE_CLOUD_PROJECT` variable in the
        Django `settings.py` file. Any project can be used, as long as the user has
        access.

        Additional requirement because this is still in dev: The OAuth is restricted
        to users registered with Pitt-Google, so contact us.

        TODO: Integrate this with Django. For now, the user interacts via command line.
        """
        # create an OAuth2Session
        client_id = settings.PITTGOOGLE_OAUTH_CLIENT_ID
        client_secret = settings.PITTGOOGLE_OAUTH_CLIENT_SECRET
        authorization_base_url = "https://accounts.google.com/o/oauth2/auth"
        redirect_uri = "https://ardent-cycling-243415.appspot.com/"  # TODO: better page
        scopes = [
            "https://www.googleapis.com/auth/logging.write",
            "https://www.googleapis.com/auth/pubsub",
        ]
        oauth2 = OAuth2Session(client_id, redirect_uri=redirect_uri, scope=scopes)

        # instruct the user to authorize
        authorization_url, state = oauth2.authorization_url(
            authorization_base_url,
            access_type="offline",
            # access_type="online",
            # prompt="select_account",
        )
        print((
            "Please visit this URL to authenticate yourself and authorize "
            "PittGoogleConsumer to make API calls on your behalf:"
            f"\n\n{authorization_url}\n"
        ))
        authorization_response = input(
            "After authorization, you should be directed to the Pitt-Google Alert "
            "Broker home page. Enter the full URL of that page (it should start with "
            "https://ardent-cycling-243415.appspot.com/):\n"
        )

        # complete the authentication
        _ = oauth2.fetch_token(
            "https://accounts.google.com/o/oauth2/token",
            authorization_response=authorization_response,
            client_secret=client_secret,
        )
        self.oauth2 = oauth2

    def stream_alerts(
        self, user_filter=None, user_callback=None, **kwargs
    ):
        """Execute a streaming pull and process alerts through the `callback`.

        The streaming pull happens in a background thread. A `queue.Queue` is used
        to communicate between threads and enforce the stopping condition(s).

        Args:
            user_filter (Callable): Used by `callback` to filter alerts before
                                    saving. It should accept a single alert as a
                                    dictionary (flat dict with fields determined by
                                    `ztf_fields` attribute).
                                    It should return the alert dict if it passes the
                                    filter, else None.

            user_callback (Callable): Used by `callback` to process alerts.
                                      It should accept a single alert as a
                                      dictionary (flat dict with fields determined by
                                      `ztf_fields` attribute).
                                      It should return True if the processing was
                                      successful; else False.

            kwargs (dict): User's parameters. Should include the parameters
                           defined in `BrokerStreamPython`'s `FilterAlertsForm`.
                           There must be at least one stopping condition
                           (`max_results` or `timeout`), else the streaming pull
                           will run forever.
        """
        # callback doesn't accept kwargs. set attribute instead.
        self.callback_kwargs = {
            "user_filter": user_filter,
            "user_callback": user_callback,
            **kwargs,
        }

        # avoid pulling down a large number of alerts that don't get processed
        flow_control = pubsub_v1.types.FlowControl(max_messages=kwargs['max_backlog'])

        # Google API has a thread scheduler that can run multiple background threads
        # and includes a queue, but I (Troy) haven't gotten it working yet.
        # self.scheduler = ThreadScheduler(ThreadPoolExecutor(max_workers))
        # self.scheduler.schedule(self.callback, lighten_alerts=lighten_alerts)

        # start pulling and processing msgs using the callback, in a background thread
        self.streaming_pull_future = self.client.subscribe(
            self.subscription_path,
            self.callback,
            flow_control=flow_control,
            # scheduler=self.scheduler,
            # await_callbacks_on_shutdown=True,
        )

        try:
            # Use the queue to count saved messages and
            # stop when we hit a max_messages or timeout stopping condition.
            num_saved = 0
            while True:
                try:
                    num_saved += self.queue.get(block=True, timeout=kwargs['timeout'])
                except Empty:
                    break
                else:
                    self.queue.task_done()
                    if kwargs['max_results'] & num_saved >= kwargs['max_results']:
                        break
            self._stop()

        except KeyboardInterrupt:
            self._stop()
            raise

        self._log_and_print(f"Saved {num_saved} messages from {self.subscription_path}")

        return self.database_list

    def _stop(self):
        """Shutdown the streaming pull in the background thread gracefully."""
        self.streaming_pull_future.cancel()  # Trigger the shutdown.
        self.streaming_pull_future.result()  # Block until the shutdown is complete.

    def callback(self, message):
        """Process a single alert; run user filter; save alert; acknowledge Pub/Sub msg.

        Used as the callback for the streaming pull.
        """
        kwargs = self.callback_kwargs

        # unpack
        try:
            alert_dict = self._unpack(message)
        except Exception as e:
            self._log_and_print(f"Error unpacking message: {e}", severity="DEBUG")
            message.nack()  # nack so message does not leave subscription
            return

        # run user filter
        if kwargs["user_filter"] is not None:
            try:
                alert_dict = kwargs["user_filter"](alert_dict, **kwargs)
            except Exception as e:
                self._log_and_print(f"Error running user_filter: {e}", severity="DEBUG")
                message.nack()
                return

        # run user callback
        if kwargs["user_callback"] is not None:
            try:
                success = kwargs["user_callback"](alert_dict)  # bool
            except Exception as e:
                success = False
                msg = f"Error running user_callback: {e}"
            else:
                if not success:
                    msg = "user_callback reported it was unsuccessful."
            finally:
                if not success:
                    self._log_and_print(msg, severity="DEBUG")
                    message.nack()
                    return

        if alert_dict is not None:
            self.save_alert(alert_dict)

            # communicate with the main thread
            self.queue.put(1)  # 1 alert successfully processed
            if kwargs['max_results'] is not None:
                # block til main thread acknowledges so we don't ack msgs that get lost
                self.queue.join()  # single background thread => one-in-one-out

        else:
            self._log_and_print("alert_dict is None")

        message.ack()

    def _unpack(self, message):
        alert_dict = avro_to_dict(message.data)
        alert_dict = self._lighten_alert(alert_dict)  # flat dict with self.ztf_fields
        alert_dict.update(self._extract_metadata(message))
        return alert_dict

    def save_alert(self, alert):
        """Save the alert to a database."""
        self.database_list.append(alert)  # fake database for demo

    def _set_ztf_fields(self, fields=None):
        """Fields to save in the `_unpack` method."""
        if fields is not None:
            self.ztf_fields = fields
        else:
            self.ztf_fields = {
                "top-level": ["objectId", "candid", ],
                "candidate": ["jd", "ra", "dec", "magpsf", "classtar", ],
                "metadata": ["message_id", "publish_time", "kafka.timestamp"]
            }

    def _extract_metadata(self, message):
        # TOM wants to serialize this and has trouble with the dates.
        # Just make everything strings for now.
        # attributes includes Kafka attributes from originating stream:
        # kafka.offset, kafka.partition, kafka.timestamp, kafka.topic
        attributes = {k: str(v) for k, v in message.attributes.items()}
        metadata = {
            "message_id": message.message_id,
            "publish_time": str(message.publish_time),
            **attributes,
        }
        return {k: v for k, v in metadata.items() if k in self.ztf_fields["metadata"]}

    def _lighten_alert(self, alert_dict):
        alert_lite = {k: alert_dict[k] for k in self.ztf_fields["top-level"]}
        alert_lite.update(
            {k: alert_dict["candidate"][k] for k in self.ztf_fields["candidate"]}
        )
        return alert_lite

    def touch_subscription(self):
        """Make sure the subscription exists and we can connect.

        If the subscription doesn't exist, try to create one (in the user's project)
        that is attached to a topic of the same name in the Pitt-Google project.

        Note that messages published before the subscription is created are not
        available.
        """
        try:
            # check if subscription exists
            sub = self.client.get_subscription(subscription=self.subscription_path)

        except NotFound:
            self._create_subscription()

        else:
            self.topic_path = sub.topic
            print(f"Subscription exists: {self.subscription_path}")
            print(f"Connected to topic: {self.topic_path}")

    def _create_subscription(self):
        """Try to create the subscription."""
        try:
            self.client.create_subscription(
                name=self.subscription_path, topic=self.topic_path
            )
        except NotFound:
            # suitable topic does not exist in the Pitt-Google project
            raise ValueError(
                (
                    f"A subscription named {self.subscription_name} does not exist"
                    "in the Google Cloud Platform project "
                    f"{settings.GOOGLE_CLOUD_PROJECT}, "
                    "and one cannot be automatically create because Pitt-Google "
                    "does not publish a public topic with the same name."
                )
            )
        else:
            self._log_and_print(f"Created subscription: {self.subscription_path}")

    def delete_subscription(self):
        """Delete the subscription.

        This is provided for the user's convenience, but it is not necessary and is not
        automatically called.

            - Storage of unacknowledged Pub/Sub messages does not result in fees.

            - Unused subscriptions automatically expire; default is 31 days.
        """
        try:
            self.client.delete_subscription(subscription=self.subscription_path)
        except NotFound:
            pass
        else:
            self._log_and_print(f'Deleted subscription: {self.subscription_path}')

    def _log_and_print(self, msg, severity="INFO"):
        self.logger.log_text(msg, severity=severity)
        print(msg)



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""Tests for the ``urls`` module."""

from django.test import TestCase
from django.urls import resolve, reverse

from broker_web.apps.alerts import urls, views


class TestUrlRouting(TestCase):
    """Test URLs are routed to the correct views"""

    app_name = urls.app_name

    def test_alerts_json_routing(self):
        """Test 'alerts-json' is routed to``AlertsJsonView``"""

        url = reverse(f'{self.app_name}:alerts-json')
        self.assertEqual(views.AlertsJsonView, resolve(url).func.view_class)

    def test_recent_alerts_routing(self):
        """Test 'recent-alerts' is routed to``RecentAlertsView``"""

        url = reverse(f'{self.app_name}:recent-alerts')
        self.assertEqual(views.RecentAlertsView, resolve(url).func.view_class)

    def test_alert_summary_routing(self):
        """Test 'alert-summary' is routed to``AlertSummaryView``"""

        dummy_alert_pk = '123'
        url = reverse(f'{self.app_name}:alert-summary', args=[dummy_alert_pk])
        self.assertEqual(views.AlertSummaryView, resolve(url).func.view_class)



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""Tests for the ``views`` module."""

from django.test import Client, TestCase
from django.urls import reverse
from django.utils.encoding import force_bytes
from django.utils.http import urlsafe_base64_encode

from broker_web.apps.signup import models, tokens, urls


class TestSignUpView(TestCase):
    """Tests for the ``SignUp`` view"""

    url_name = f'{urls.app_name}:signup'
    template = 'signup/create_new_user.html'

    def setUp(self):
        self.client = Client()

    def test_get(self):
        """Test ``get`` method returns correct template and object id"""

        url = reverse(self.url_name)
        response = self.client.get(url)

        self.assertEqual(200, response.status_code)
        self.assertTemplateUsed(response, self.template)


class TestActivateAccount(TestCase):
    """Tests for the ``ActivateAccount`` view"""

    url_name = f'{urls.app_name}:activate'
    invalid_token_template = 'signup/invalid_activation_link.html'
    valid_token_template = 'signup/activation_success.html'

    def setUp(self):
        # Create client and test user
        self.client = Client()
        self.test_user = models.CustomUser(pk=1)

        # Create activation token for custom user
        self.uidb64 = urlsafe_base64_encode(force_bytes(self.test_user.pk))
        self.test_token = tokens.account_activation_token.make_token(self.test_user)

    def test_get_with_invalid_token(self):
        """Test get request with invalid token redirects to invalid token page"""

        # noinspection SpellCheckingInspection
        bad_signup_token = {'uidb64': 'AB', 'token': 'CDE-FGHIJK'}
        url = reverse(self.url_name, kwargs=bad_signup_token)
        response = self.client.get(url)

        self.assertEqual(200, response.status_code)
        self.assertTemplateUsed(response, self.invalid_token_template)

    def test_get_with_valid_token(self):
        """Test get request with invalid token redirects to success page"""

        good_signup_token = {'uidb64': self.uidb64, 'token': self.test_token}
        url = reverse(self.url_name, kwargs=good_signup_token)
        response = self.client.get(url)

        self.assertEqual(200, response.status_code)
        self.assertTemplateUsed(response, self.valid_token_template)



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""Custom token generation for user signup tasks"""

from django.contrib.auth.tokens import PasswordResetTokenGenerator


class TokenGenerator(PasswordResetTokenGenerator):
    def _make_hash_value(self, user, timestamp):
        return user.pk + timestamp + user.is_active


account_activation_token = TokenGenerator()



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""The ``views`` module defines ``View`` objects for converting web requests
into rendered responses.

.. autosummary::
   :nosignatures:

   broker_web.apps.getting_started.views.Introduction
   broker_web.apps.getting_started.views.DataProducts
   broker_web.apps.getting_started.views.TechnicalResources
   broker_web.apps.getting_started.views.DataAccess
   broker_web.apps.getting_started.views.BrokerDesign
"""

from django.views.generic import TemplateView

dir_name = 'getting_started/'

Introduction = TemplateView.as_view(
    template_name=dir_name + 'introduction.html')

DataProducts = TemplateView.as_view(
    template_name=dir_name + 'data_products.html')

TechnicalResources = TemplateView.as_view(
    template_name=dir_name + 'technical_resources.html')

DataAccess = TemplateView.as_view(
    template_name=dir_name + 'data_access.html')

BrokerDesign = TemplateView.as_view(
    template_name=dir_name + 'broker_design.html')

for view in (Introduction, DataProducts, TechnicalResources, DataAccess, BrokerDesign):
    view.__doc__ = 'Static template view'



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""Custom django applications built for the Pitt-Google Broker website"""

# Todo apps:
# - The ``alerts`` and ``objects`` apps are very similar, but I expect them
#   to diverge during development. If this is incorrect, consider merging them.



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""A django application for serving data concerning recently observed
astronomical objects. The focus of this app is on the object itself, and not
the corresponding alerts.
"""

# Todo-objects:
# - Add / use PubSub configuration to / from global settings
# - Populate the ObjectsJsonView view using PubSub messages
# - Use arguments from the filter results from to filter values in ObjectsJsonView
# - Move ``forms.topics`` into global settings



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""Tests for the ``forms`` module."""

from django.test import TestCase

from broker_web.apps.objects.forms import FilterObjectsForm


class TestFilterObjectsForm(TestCase):
    """Test the ``FilterObjectsForm`` validates correctly"""

    def test_empty_form(self):
        """Test that an empty form is valid"""

        form = FilterObjectsForm(data={})
        self.assertTrue(form.is_valid())



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""Tests for the ``urls`` module."""

from django.test import TestCase
from django.urls import resolve, reverse

from broker_web.apps.signup import urls, views


class TestUrlRouting(TestCase):
    """Test URLs are routed to the correct views"""

    app_name = urls.app_name

    def test_signup_routing(self):
        """Test 'signup' is routed to``SignUp``"""

        url = reverse(f'{self.app_name}:signup')
        self.assertEqual(views.SignUp, resolve(url).func.view_class)

    def test_activation_sent_routing(self):
        """Test 'activation-sent' is routed to``activation_sent_view``"""

        url = reverse(f'{self.app_name}:activation-sent')
        self.assertEqual(views.ActivationSentView, resolve(url).func)

    def test_activate_routing(self):
        """Test 'activate' is routed to``ActivateAccount``"""

        # noinspection SpellCheckingInspection
        dummy_activation_key = {'uidb64': 'AB', 'token': 'CDE-FGHIJK'}
        url = reverse(f'{self.app_name}:activate', kwargs=dummy_activation_key)
        self.assertEqual(views.ActivateAccount, resolve(url).func.view_class)



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""General utilities used across apps

Paginating JSON Responses
-------------------------

The ``paginate_to_json`` function handles the generation of JSON responses to
HTTP requests for pagination of data. For example, it can be used in a view as:

.. code-block:: python
   :linenos:

   from django.views.generic import View

   from apps.utils import paginate_to_json

   class MyView(View):

       template = 'myt_template_path.html'

       def get(self, request, *args, **kwargs):
           data = [{'field1': 'value1'}, {'field1': 'value2'}, ...]
           return paginate_to_json(request, data)
"""

from django.http import JsonResponse


def paginate_to_json(request, data):
    """Paginate a list of dicts and return as a ``JsonResponse``

    For expected in/outputs of paginated data, see
    https://datatables.net/manual/server-side .

    Args:
        request (HttpRequest): Incoming HTTP request
        data           (list): The data to paginate
    """

    start = int(request.GET.get('start', 0))
    length = int(request.GET.get('length', 10))
    draw = request.GET.get('draw', -1)
    limit = request.GET.get('limit', None)

    if limit:
        limit = int(limit)

    # Paginate data
    data = data[:limit]
    paginated_alerts = data[start:start + length]

    response = {
        'draw': draw,
        'data': paginated_alerts,
        'recordsTotal': len(data),
        'recordsFiltered': len(data),
    }

    return JsonResponse(response)



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""A django application for managing user profiles and topic subscriptions"""

# Todo-subscriptions:
# - Add a model to which users can manage their topic subscriptions



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""The ``forms`` module defines views forms for data entry and query
construction.

.. autosummary::
   :nosignatures:

   broker_web.apps.contact.forms.ContactForm
"""

from django import forms


class ContactForm(forms.Form):
    """Form to send a "contact us" email to the maintainers"""

    email = forms.EmailField(required=True)
    subject = forms.CharField(required=True)
    message = forms.CharField(widget=forms.Textarea, required=True)



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""Tests for the ``forms`` module."""

from django.test import TestCase

from broker_web.apps.signup.forms import CustomUserChangeForm, CustomUserCreationForm


class TestCustomUserCreationForm(TestCase):
    """Test the ``CustomUserCreationForm`` validates correctly"""

    def test_empty_form(self):
        """Test that an empty form is invalid"""

        form = CustomUserCreationForm(data={})
        self.assertFalse(form.is_valid())


class TestCustomUserChangeForm(TestCase):
    """Test the ``CustomUserChangeForm`` validates correctly"""

    def test_empty_form(self):
        """Test that an empty form is invalid"""

        form = CustomUserChangeForm(data={})
        self.assertFalse(form.is_valid())


# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""Tests for the ``views`` module."""

from django.test import Client, TestCase
from django.urls import reverse

from broker_web.apps.getting_started import urls


class TestTemplates(TestCase):
    """Test views use the correct templates"""

    app_name = urls.app_name

    def setUp(self):
        self.client = Client()

    def test_introduction_template(self):
        """Test ``Introduction`` view uses the correct template"""

        url = reverse(f'{self.app_name}:introduction')
        response = self.client.get(url)

        self.assertEqual(200, response.status_code)
        self.assertTemplateUsed(response, 'getting_started/introduction.html')

    def test_data_products_template(self):
        """Test ``DataProducts`` view uses the correct template"""

        url = reverse(f'{self.app_name}:data-products')
        response = self.client.get(url)

        self.assertEqual(200, response.status_code)
        self.assertTemplateUsed(response, 'getting_started/data_products.html')

    def test_technical_resources_template(self):
        """Test ``TechnicalResources`` view uses the correct template"""

        url = reverse(f'{self.app_name}:technical-resources')
        response = self.client.get(url)

        self.assertEqual(200, response.status_code)
        self.assertTemplateUsed(response, 'getting_started/technical_resources.html')

    def test_data_access_template(self):
        """Test ``DataAccess`` view uses the correct template"""

        url = reverse(f'{self.app_name}:data-access')
        response = self.client.get(url)

        self.assertEqual(200, response.status_code)
        self.assertTemplateUsed(response, 'getting_started/data_access.html')

    def test_broker_design_template(self):
        """Test ``BrokerDesign`` view uses the correct template"""

        url = reverse(f'{self.app_name}:broker-design')
        response = self.client.get(url)

        self.assertEqual(200, response.status_code)
        self.assertTemplateUsed(response, 'getting_started/broker_design.html')



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""The ``views`` module defines ``View`` objects for converting web requests
into rendered responses.

.. autosummary::
   :nosignatures:

   broker_web.apps.signup.views.ActivateAccount
   broker_web.apps.signup.views.ActivationSentView
   broker_web.apps.signup.views.SignUp
"""

from django.contrib.sites.models import Site
from django.core.mail import EmailMessage
from django.shortcuts import render
from django.template.loader import render_to_string
from django.urls import reverse_lazy
from django.utils.encoding import force_bytes
from django.utils.http import urlsafe_base64_decode, urlsafe_base64_encode
from django.views.generic import CreateView, View
from django.views.generic import TemplateView

from .forms import CustomUserCreationForm
from .models import CustomUser
from .tokens import account_activation_token

ActivationSentView = TemplateView.as_view(template_name='signup/activation_link_sent.html')


class SignUp(CreateView):
    """View that handles ``CustomUser`` creation"""

    template_name = 'signup/create_new_user.html'
    form_class = CustomUserCreationForm

    # Todo: Include app name in reverse lookup in case app is namespaced differently
    success_url = reverse_lazy('signup:activation-sent')

    def form_valid(self, form):
        """Sends email confirmation for new user creation

        Called after form is validated

        Args:
            form (django.forms.Form): User creation form
        """

        # Create an inactive user
        user = form.save(commit=False)
        user.is_active = False
        user.save()

        current_site = Site.objects.get_current()

        email_subject = 'Activate Your Account'
        message = render_to_string('signup/activate_account.html', {
            'user': user,
            'domain': current_site.domain,
            'uid': urlsafe_base64_encode(force_bytes(user.pk)),
            'token': account_activation_token.make_token(user),
        })

        to_email = form.cleaned_data.get('email')
        email = EmailMessage(email_subject, message, to=[to_email])
        email.send()

        # Parent class ``form_valid`` redirects to ``self.success_url``
        return super().form_valid(form)


class ActivateAccount(View):
    """View that account verification"""

    def get(self, request, uidb64, token):
        """Handle an incoming HTTP request

        Args:
            request (HttpRequest): Incoming HTTP request
            uidb64          (str): Base 64 encoded user id
            token           (str): Activationn auth token

        Returns:
            Outgoing HTTPResponse
        """

        try:
            uid = force_bytes(urlsafe_base64_decode(uidb64))
            user = CustomUser.objects.get(pk=uid)

        except(TypeError, ValueError, OverflowError, CustomUser.DoesNotExist):
            user = None

        if user is not None and account_activation_token.check_token(user, token):
            user.is_active = True
            user.save()
            return render(request, 'signup/activation_success.html')

        else:
            return render(request, 'signup/invalid_activation_link.html')



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""The ``views`` module defines ``View`` objects for converting web requests
into rendered responses.

.. autosummary::
   :nosignatures:

   broker_web.apps.alerts.views.AlertsJsonView
   broker_web.apps.alerts.views.AlertSummaryView
   broker_web.apps.alerts.views.RecentAlertsView
"""

import os

from django.conf import settings
from django.shortcuts import render
from django.views.generic import View
from google.api_core.exceptions import BadRequest
from google.cloud import bigquery

from .forms import FilterAlertsForm
from ..utils import paginate_to_json
from ..utils.templatetags.utility_tags import jd_to_readable_date

NUM_ALERTS = 10_000

if 'BUILD_IN_RTD' not in os.environ:
    CLIENT = bigquery.Client()


class AlertsJsonView(View):
    """Serves recent alerts as a paginated JSON response"""

    @staticmethod
    def fetch_alerts_as_dicts(request, num_alerts=NUM_ALERTS):
        """Returns a list of recent alerts messages as dicts

        Args:
            request (HttpRequest): Incoming HTTP request
            num_alerts      (int): Maximum number of alerts to return

        Return:
            A list of dictionaries representing
        """

        # Select top most recent alerts
        query = CLIENT.query(f"""
            SELECT
                publisher, 
                CAST(candidate.candid AS STRING) as alert_id,
                CASE candidate.fid WHEN 1 THEN 'g' WHEN 2 THEN 'R' WHEN 3 THEN 'i' END as filter,
                ROUND(candidate.magpsf, 2) as magnitude,
                objectId as object_id, 
                candidate.jd as pub_time, 
                ROUND(candidate.ra, 2) as ra, 
                ROUND(candidate.dec, 2) as dec 
            FROM `{settings.ZTF_ALERTS_TABLE_NAME}` 
            ORDER BY pub_time DESC
            LIMIT {num_alerts}
        """)

        output = []
        for row in query.result():
            row = dict(row)
            row['pub_time'] = jd_to_readable_date(row['pub_time'])
            output.append(row)

        return output

    def get(self, request):
        """Handle an incoming HTTP request

        Args:
            request (HttpRequest): Incoming HTTP request

        Returns:
            Outgoing JsonResponse
        """

        alerts = self.fetch_alerts_as_dicts(request)
        return paginate_to_json(request, alerts)


class RecentAlertsView(View):
    """Provides a summary table of recently ingested alerts"""

    template = 'alerts/recent-alerts.html'

    def get(self, request):
        """Handle an incoming HTTP request

        Args:
            request (HttpRequest): Incoming HTTP request

        Returns:
            Outgoing HTTPResponse
        """

        context = {'form': FilterAlertsForm()}
        return render(request, self.template, context)

    def post(self, request):  # Todo: Add filtering from form and update tests
        """Fill in the page's form with values from the POST request

        Args:
            request (HttpRequest): Incoming HTTP request

        Returns:
            Outgoing HTTPResponse
        """

        form = FilterAlertsForm(request.POST)
        return render(request, self.template, {'form': form})


class AlertSummaryView(View):
    """Displays information about a single alert"""

    template = 'alerts/alert_summary.html'

    @staticmethod
    def _get_ztf_alert_data(alert_id):
        """Retrieve alert data for a ZTF alert ID

        Args:
            alert_id (int): Id of the alert to retrieve data for

        Return:
            A dictionary of alert data if available else None
        """

        query = CLIENT.query(f"""
            SELECT 
                schemavsn, publisher, objectId, candid, candidate, 
                cutoutScience.stampData as cutout_science, 
                cutoutTemplate.stampData as cutout_template, 
                cutoutDifference.stampData as cutout_difference
            FROM `{settings.ZTF_ALERTS_TABLE_NAME}` 
            WHERE candidate.candid={alert_id}
        """)

        try:
            query_result = query.result()

        except BadRequest:  # Alert id was likely a string and thus invalid
            return None

        # Return first value from the iterable
        for row in query_result:
            out_data = dict()
            for k, v in row.items():
                if isinstance(v, dict):  # Support for nested dictionaries
                    out_data.update(v)

                else:
                    out_data[k] = v

            return out_data

    def get_alert_data_for_id(self, alert_id, survey):
        """Retrieve alert data for a given alert ID

        Args:
            alert_id (int): Id of the alert to retrieve data for
            survey   (str): Parent survey of the alert

        Return:
            A dictionary of alert data
        """

        if survey.lower() == 'ztf':
            return self._get_ztf_alert_data(alert_id)

        raise NotImplementedError(f'Database alert queries not implemented for survey {survey}')

    @staticmethod
    def get_value_added_data_for_id(alert_id, survey):
        """Retrieve value added data products for a given alert ID

        Args:
            alert_id (int): Id of the alert to retrieve data for
            survey   (str): Parent survey of the alert

        Return:
            A dictionary of value added data products
        """

        # Todo: This function can probably be broken into multiple
        # functions based on different data products. It depends
        # how the back / front end is designed
        return dict()

    def get(self, request, *args, **kwargs):
        """Handle an incoming HTTP request

        Args:
            request (HttpRequest): Incoming HTTP request

        Returns:
            Outgoing JsonResponse
        """

        alert_id = kwargs['pk']
        survey = kwargs.get('survey', 'ztf')
        alert_data = self.get_alert_data_for_id(alert_id, survey)
        if alert_data is None:
            return render(request, 'alerts/error_404.html', {'alert_id': alert_id})

        context = {
            'alert_data': alert_data,
            'alert_id': alert_id,
            'survey': survey,
            'science_image': alert_data.pop('cutout_science'),
            'template_image': alert_data.pop('cutout_template'),
            'difference_image': alert_data.pop('cutout_difference')
        }

        return render(request, self.template, context)



# # !/usr/bin/env python3
# # -*- coding: UTF-8 -*-
#
# """Django settings
#
# For more information on this file, see
# https://api_docs.djangoproject.com/en/3.0/topics/settings/
#
# For the full list of settings and their values, see
# https://api_docs.djangoproject.com/en/3.0/ref/settings/
# """
#
# import os
# from warnings import warn
#
# import environ
# import pymysql
#
# BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
# pymysql.version_info = (9, 9, 99, "final", 0)  # https://stackoverflow.com/a/59591269
# pymysql.install_as_MySQLdb()
# env = environ.Env()
#
# # Read environment settings from file only if we are not deployed to App Engine
# running_in_app_engine = os.getenv('GAE_APPLICATION', False)
# if not running_in_app_engine:
#     environ.Env.read_env(os.path.join(os.path.dirname(BASE_DIR), '.env'))
#
# # SECURITY WARNING: Make sure the following settings are properly configured in production
# # App Engine's security features ensure that it is safe to
# # have ALLOWED_HOSTS = ['*'] when the app is deployed. If you deploy a Django
# # app not on App Engine, make sure to set an appropriate host here.
# # See https://docs.djangoproject.com/en/1.10/ref/settings/
# ###############################################################################
# SECRET_KEY = env.str('SECRET_KEY')
# DEBUG = env.bool('DEBUG', default=False)
# ALLOWED_HOSTS = env.list('ALLOWED_HOSTS', default=['*'])
# EMAIL_BACKEND = 'django.core.mail.backends.console.EmailBackend'
# CONTACT_EMAILS = env.list('CONTACT_EMAILS', default=[])
# RECAPTCHA_PUBLIC_KEY = env.str('RECAPTCHA_PUBLIC_KEY', default='')
# RECAPTCHA_PRIVATE_KEY = env.str('RECAPTCHA_PRIVATE_KEY', default='')
# ZTF_ALERTS_TABLE_NAME = env.str('ZTF_ALERTS_TABLE_NAME', default='ardent-cycling-243415.ztf_alerts.alerts')
# ZTF_SALT2_TABLE_NAME = env.str('ZTF_SALT2_TABLE_NAME', default='ardent-cycling-243415.ztf_alerts.salt2')
# ZTF_SALT2_IMAGE_BUCKET = env.str('ZTF_SALT2_TABLE_NAME', default='ardent-cycling-243415_ztf-sncosmo')
# NOCAPTCHA = True
# ###############################################################################
#
# if not (RECAPTCHA_PUBLIC_KEY or RECAPTCHA_PRIVATE_KEY):
#     warn('Recaptcha keys not set in environment. '
#          'Recaptcha verification may not work correctly')
#
# INSTALLED_APPS = [
#     'django.contrib.admin',  # Django administration interface
#     'django.contrib.auth',  # Core authentication framework and associated models
#     'django.contrib.contenttypes',  # Allows permissions to be associated with models
#     'django.contrib.sessions',  # Session framework for cookies handling
#     'django.contrib.messages',  # Displays one-time notification messages
#     'django.contrib.staticfiles',  # Renders paths to static files
#     'django.contrib.sites',  # Handles multi-site hosting on multiple domains
#     'django_extensions',  # Extends capability of django management commands
#     'guardian',  # Extra authentication backend with per object permissions
#     'bootstrap4',  # Front-end component library for building templates
#     'crispy_forms',  # Makes forms look pretty
#     'captcha',  # Implements Google recaptcha service
#
#     # Custom apps
#     'broker_web.apps.alerts',  # Displays alert information
#     'broker_web.apps.contact',  # Adds a "contact Us" form
#     'broker_web.apps.getting_started',  # Handles the getting started guide
#     'broker_web.apps.objects',  # Displays object information
#     'broker_web.apps.signup',  # Handles user creation / authentication
#     'broker_web.apps.subscriptions',  # Handles alert topic subscriptions
#     'broker_web.apps.utils',  # Handles alert topic subscriptions
# ]
#
# # App configuration
# CRISPY_TEMPLATE_PACK = 'bootstrap4'
#
# # Site configuration
# ROOT_URLCONF = 'broker_web.main.urls'
# SITE_ID = 1  # For description, see https://stackoverflow.com/a/25468782/6466457
# LOGIN_REDIRECT_URL = '/'
# LOGIN_URL = '/users/login/'
# AUTH_USER_MODEL = 'signup.CustomUser'  # Use custom user model for authentication
#
# MIDDLEWARE = [
#     'django.middleware.security.SecurityMiddleware',
#     'django.contrib.sessions.middleware.SessionMiddleware',
#     'django.middleware.common.CommonMiddleware',
#     'django.middleware.csrf.CsrfViewMiddleware',
#     'django.contrib.auth.middleware.AuthenticationMiddleware',
#     'django.contrib.messages.middleware.MessageMiddleware',
#     'django.middleware.clickjacking.XFrameOptionsMiddleware'
# ]
#
# TEMPLATES = [
#     {
#         'BACKEND': 'django.template.backends.django.DjangoTemplates',
#         'DIRS': [os.path.join(BASE_DIR, 'templates')],
#         'APP_DIRS': True,
#         'OPTIONS': {
#             'context_processors': [
#                 'django.template.context_processors.debug',
#                 'django.template.context_processors.request',
#                 'django.contrib.auth.context_processors.auth',
#                 'django.contrib.messages.context_processors.messages',
#             ],
#         },
#     },
# ]
#
# # Database connection settings
# default_db_name = 'web_backend'
# if running_in_app_engine:
#     # Running on production App Engine, so connect to Google Cloud SQL using
#     # the unix socket at /cloudsql/<your-cloudsql-connection string>
#     DATABASES = {
#         'default': {
#             'ENGINE': 'django.db.backends.mysql',
#             'HOST': '/cloudsql/ardent-cycling-243415:us-east1:broker-web',
#             'USER': env.str('DB_USER'),
#             'PASSWORD': env.str('DB_PASSWORD'),
#             'NAME': env.str('DB_NAME', default_db_name),
#         }
#     }
#
# elif os.getenv('GAE_REMOTE', False):
#     # Running locally, but connect to Cloud SQL via the proxy.
#     # To start the proxy see https://cloud.google.com/sql/docs/mysql-connect-proxy
#     DATABASES = {
#         'default': {
#             'ENGINE': 'django.db.backends.mysql',
#             'HOST': '127.0.0.1',
#             'PORT': '3306',
#             'USER': env.str('DB_USER'),
#             'PASSWORD': env.str('DB_PASSWORD'),
#             'NAME': env.str('DB_NAME', default_db_name),
#             'TEST_NAME': f'test_{env.str("DB_NAME", default_db_name)}'
#         }
#     }
#
# else:
#     # Running against a local db
#     DATABASES = {
#         'default': {
#             'ENGINE': 'django.db.backends.mysql',
#             'NAME': env.str('DB_NAME', default_db_name),
#             'USER': env.str('DB_USER'),
#             'PASSWORD': env.str('DB_PASSWORD'),
#             'HOST': env.str('DB_HOST', ''),
#             'PORT': env.str('DB_PORT', ''),
#             'TEST_NAME': f'test_{env.str("DB_NAME", default_db_name)}'
#         }
#     }
#
# # Password validation
# # https://docs.djangoproject.com/en/3.0/ref/settings/#auth-password-validators
# AUTH_PASSWORD_VALIDATORS = [
#     {'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator'},
#     {'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator'},
#     {'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator'},
#     {'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator'},
# ]
#
# # Authentication configuration
# AUTHENTICATION_BACKENDS = (
#     'django.contrib.auth.backends.ModelBackend',
#     'guardian.backends.ObjectPermissionBackend',
# )
#
# # Internationalization
# # https://docs.djangoproject.com/en/3.0/topics/i18n/
# LANGUAGE_CODE = 'en-us'
# TIME_ZONE = 'UTC'
# USE_I18N = True
# USE_L10N = False
# USE_TZ = True
# DATETIME_FORMAT = 'Y-m-d H:m:s'
# DATE_FORMAT = 'Y-m-d'
#
# # Static files (CSS, JavaScript, Images, etc.)
# # https://docs.djangoproject.com/en/3.0/howto/static-files/
# STATICFILES_DIRS = [os.path.join(BASE_DIR, 'static')]
# if running_in_app_engine:
#     STATIC_URL = 'https://storage.googleapis.com/broker-web-static/static/'
#
# else:
#     STATIC_URL = '/static/'



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""Tests for the ``urls`` module."""

from django.test import TestCase

from broker_web.apps.signup import tokens
from broker_web.apps.signup.models import CustomUser


class TestTokenGenerator(TestCase):
    """Tests for the account activation token generator"""

    def setUp(self):
        """Create a test user to generate a token for"""

        # pk is used when calculating the token and so must be set
        self.test_user = CustomUser(pk=1)
        self.token_generator = tokens.TokenGenerator()

    def test_valid_token(self):
        """Test generated token is identified as invalid"""

        valid_token = self.token_generator.make_token(self.test_user)
        is_valid = self.token_generator.check_token(self.test_user, valid_token)
        self.assertTrue(is_valid)

    def test_invalid_token(self):
        """Test dummy token is identified as invalid"""

        invalid_token = 'not a token'
        is_valid = self.token_generator.check_token(self.test_user, invalid_token)
        self.assertFalse(is_valid)

    def test_user_token_is_unique(self):
        """Test tokens are unique even when generated for different users"""

        token1 = self.token_generator.make_token(self.test_user)

        self.test_user.pk += 10
        token2 = self.token_generator.make_token(self.test_user)

        self.assertNotEqual(token1, token2)



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""The ``views`` module defines ``View`` objects for converting web requests
into rendered responses.
"""

from django.shortcuts import render


def error_404(request, exception):
    return render(request, 'error_pages/error_404.html')


def error_500(request):
    return render(request, 'error_pages/error_500.html')



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""Defines views for converting a Web requests into a Web responses"""

from django.shortcuts import render
from django.views.generic import TemplateView, View

why_pgb_view = TemplateView.as_view(template_name='broker_web/why_pgb.html')


class IndexView(View):
    """View for the index page"""

    def get(self, request):
        """Handle an incoming HTTP request

        Args:
            request (HttpRequest): Incoming HTTP request

        Returns:
            Outgoing HTTPResponse
        """

        context = {
            'online': False
        }

        return render(request, 'broker_web/index.html', context)



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""TOM Toolkit broker to query a BigQuery table via the Python API.

Relies on `ConsumerDatabasePython` to manage the connections and work with data.

See especially:

.. autosummary::
   :nosignatures:

   BrokerDatabasePython.request_alerts
"""

from django import forms
from tom_alerts.alerts import GenericQueryForm, GenericAlert, GenericBroker

from .consumer_database_python import ConsumerDatabasePython
from .utils.templatetags.utility_tags import jd_to_readable_date


class FilterAlertsForm(GenericQueryForm):
    """Basic form for filtering alerts; currently implemented in the SQL statement.

    Fields:
        objectId (``CharField``)

        candid (``IntegerField``)

        max_results (``IntegerField``)
    """

    objectId = forms.CharField(required=False)
    candid = forms.IntegerField(required=False)
    max_results = forms.IntegerField(
        required=True, initial=100, min_value=1
    )


class BrokerDatabasePython(GenericBroker):
    """Pitt-Google broker to query alerts from the database via the Python client.

    Base class: ``tom_alerts.alerts.GenericBroker``
    """

    name = "Pitt-Google DatabasePython"
    form = FilterAlertsForm

    def fetch_alerts(self, parameters):
        """Entry point to query and filter alerts."""
        clean_params = self._clean_parameters(parameters)

        table_name = "ztf_alerts.alerts"
        self.consumer = ConsumerDatabasePython(table_name)

        alerts = self.request_alerts(clean_params)

        return iter(alerts)

    def request_alerts(self, parameters):
        """Query alerts using the user filter and unpack.

        The SQL statement returned by the `Consumer` implements the current user filter.

        Returns:
            alerts (List[dict])
        """
        sql_stmnt, job_config = self.consumer.create_sql_stmnt(parameters)
        query_job = self.consumer.client.query(sql_stmnt, job_config=job_config)
        alerts = self.consumer.unpack_query(query_job)  # List[dict]
        return alerts

    def _clean_parameters(self, parameters):
        clean_params = dict(parameters)

        # make sure objectId and candid are not both set
        if (len(clean_params["objectId"]) > 0) & (clean_params["candid"] is not None):
            raise forms.ValidationError(
                "Only one of either objectId or candid can be used to filter."
            )

        return clean_params

    def to_generic_alert(self, alert):
        """Map the Pitt-Google alert to a TOM `GenericAlert`."""
        return GenericAlert(
            timestamp=jd_to_readable_date(alert["jd"]),
            url=self.consumer.query_url,
            id=alert["candid"],
            name=alert["objectId"],
            ra=alert["ra"],
            dec=alert["dec"],
            mag=alert["mag"],
            score=alert["classtar"],
        )



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""The ``models`` module defines ``Model`` objects for representing backend DB
constructs.
"""

from django.contrib.auth.models import AbstractBaseUser
from django.contrib.auth.models import PermissionsMixin
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from .managers import CustomUserManager


class CustomUser(AbstractBaseUser, PermissionsMixin):
    """Custom user model for authentication"""

    email = models.EmailField(_('email address'), unique=True)
    first_name = models.CharField(max_length=100)
    last_name = models.CharField(max_length=100)
    country = models.CharField(max_length=100)
    affiliation = models.CharField(max_length=100)
    date_joined = models.DateTimeField(default=timezone.now)
    is_staff = models.BooleanField(default=False)
    is_active = models.BooleanField(default=True)

    USERNAME_FIELD = 'email'
    REQUIRED_FIELDS = []

    objects = CustomUserManager()

    def __repr__(self):
        return f'<User(email={self.email})>'



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""The `tom_pittgoogle` module contains Pitt-Google's TOM broker implementations."""

# import django
from django.core.wsgi import get_wsgi_application
import os


if 'BUILD_IN_RTD' in os.environ:
    os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'tom_pittgoogle.settings')
    # django.setup()
    application = get_wsgi_application()



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""The ``views`` module defines ``View`` objects for converting web requests
into rendered responses.

.. autosummary::
   :nosignatures:

   broker_web.apps.objects.views.ObjectsJsonView
   broker_web.apps.objects.views.RecentAlertsJsonView
   broker_web.apps.objects.views.ObjectSummaryView
   broker_web.apps.objects.views.RecentObjectsView
"""

import os

from django.conf import settings
from django.shortcuts import render
from django.views.generic import View
from google.cloud import bigquery

from .forms import FilterObjectsForm
from ..utils import paginate_to_json
from ..utils.templatetags.utility_tags import jd_to_readable_date

if 'BUILD_IN_RTD' not in os.environ:
    CLIENT = bigquery.Client()


###############################################################################
# JSON views
###############################################################################

class RecentObjectsJsonView(View):
    """View for serving recently observed objects as a paginated JSON response"""

    @staticmethod
    def fetch_objects(limit=10_000):
        """Returns a list of objects with recently issued alerts as a list of dicts

        Args:
            limit (int): Maximum number of alerts to return

        Return:
            A list of dictionaries representing
        """

        # Select the most recent alert for each object
        query = CLIENT.query(f"""
            SELECT 
                DISTINCT objectId as object_id, 
                publisher,
                CAST(candidate.candid AS STRING) recent_alert_id, 
                candidate.jd as pub_time,
                ARRAY_LENGTH( prv_candidates ) as num_alerts,
                ROUND(candidate.ra, 2) as ra, 
                ROUND(candidate.dec, 2) as dec
            FROM `{settings.ZTF_ALERTS_TABLE_NAME}`
            ORDER BY pub_time DESC
            LIMIT {limit}
           """)

        output = []
        for row in query.result():
            row = dict(row)
            row['pub_time'] = jd_to_readable_date(row['pub_time'])
            output.append(row)

        return output

    def get(self, request):
        """Handle an incoming HTTP request

        Args:
            request (HttpRequest): Incoming HTTP request

        Returns:
            Outgoing JsonResponse
        """

        # Get all available messages
        objects = self.fetch_objects()
        return paginate_to_json(request, objects)


class RecentObjectAlertsJsonView(View):
    """JSON rendering of recent alerts for a given object"""

    @staticmethod
    def fetch_object_alerts(object_id, limit=50):
        """Return a list of all alerts corresponding to an object Id

        Args:
            object_id (str): Object identifier
            limit     (int): Maximum number of alerts to return

        Returns:
            A list of dictionaries
        """

        # Select all alerts for the given object
        query = CLIENT.query(f"""
            SELECT 
                 publisher,
                 candidate.jd as pub_time,
                 CAST(candidate.candid AS STRING) as alert_id,
                 CASE candidate.fid WHEN 1 THEN 'g' WHEN 2 THEN 'R' WHEN 3 THEN 'i' END as filter,
                 ROUND(candidate.magpsf, 2) as magnitude
            FROM `{settings.ZTF_ALERTS_TABLE_NAME}`
            WHERE objectId="{object_id}"
            LIMIT {limit}
        """)

        out_data = []
        for row in query.result():
            row_dict = dict(row)
            row_dict['jd'] = row_dict['pub_time']
            row_dict['pub_time'] = jd_to_readable_date(row_dict['pub_time'])
            out_data.append(row_dict)

        return out_data

    def get(self, request, *args, **kwargs):
        """Handle an incoming HTTP request

        Args:
            request (HttpRequest): Incoming HTTP request

        Returns:
            Outgoing JsonResponse
        """

        # Get all available messages
        alerts = self.fetch_object_alerts(kwargs['pk'])
        return paginate_to_json(request, alerts)


class Salt2FitsJsonView(View):
    """View for serving recent Salt2 fit results as a paginated JSON response"""

    @staticmethod
    def fetch_salt2_fits(object_id=None, limit=1000):
        """Return a list of recent Salt2 fits for a given astronomical object

        Args:
            object_id (str): Object identifier
            limit     (int): Maximum number of fits to return

        Returns:
            A list of dictionaries
        """

        condition = 'success=1'
        if object_id:  # Select all alerts for the given object
            condition += f' AND objectId="{object_id}"'

        query = CLIENT.query(f"""
            SELECT
                objectId as object_id,
                CAST(candId AS STRING) as alert_id,  
                ROUND(chisq, 2) as chisq, 
                ndof,
                ROUND(z, 4) as z, 
                ROUND(z_err, 6) as z_err,
                ROUND(t0, 2) as t0, 
                ROUND(t0_err, 4) as t0_err,
                ROUND(x0, 6) as x0, 
                ROUND(x0_err, 8) as x0_err,      
                ROUND(x1, 2) as x1, 
                ROUND(x1_err, 4) as x1_err,     
                ROUND(c, 2) as c, 
                ROUND(c_err, 4) as c_err,
            FROM  `{settings.ZTF_SALT2_TABLE_NAME}`
            WHERE {condition}
            LIMIT {limit}
        """)

        return [dict(row) for row in query.result()]

    def get(self, request, *args, **kwargs):
        """Handle an incoming HTTP request

        Args:
            request (HttpRequest): Incoming HTTP request

        Returns:
            Outgoing JsonResponse
        """

        json_data = self.fetch_salt2_fits(object_id=kwargs.get('pk', None))
        return paginate_to_json(request, json_data)


###############################################################################
# Views for individual web pages
###############################################################################

class RecentObjectsView(View):
    """View for displaying a summary table of objects with recent alerts"""

    template = 'objects/recent_objects.html'

    def get(self, request):
        """Handle an incoming HTTP request

        Args:
            request (HttpRequest): Incoming HTTP request

        Returns:
            Outgoing HTTPResponse
        """

        context = {'form': FilterObjectsForm()}
        return render(request, self.template, context)

    def post(self, request):
        """Fill in the page's form with values from the POST request

        Args:
            request (HttpRequest): Incoming HTTP request

        Returns:
            Outgoing HTTPResponse
        """

        form = FilterObjectsForm(request.POST)
        return render(request, self.template, {'form': form})


class ObjectSummaryView(View):
    """View for displaying a table of all recent objects matching a query"""

    template = 'objects/object_summary.html'

    def get(self, request, *args, **kwargs):
        """Handle an incoming HTTP request

        Args:
            request (HttpRequest): Incoming HTTP request

        Returns:
            Outgoing JsonResponse
        """

        return render(request, self.template, context={
            'object_id': kwargs['pk'],
            'salt2_image_bucket': settings.ZTF_SALT2_IMAGE_BUCKET})


class Salt2FitView(View):
    """View for displaying a table of salt2 fits to recently observed objects"""

    template = 'objects/salt2_fits.html'

    def get(self, request, *args, **kwargs):
        """Handle an incoming HTTP request

        Args:
            request (HttpRequest): Incoming HTTP request

        Returns:
            Outgoing JsonResponse
        """

        return render(request, self.template)



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""The ``utility_tags`` module defines functions for casting between
different data types and formats. These functions are automatically
registered as template tags for use in the Django template engine.

.. autosummary::
   :nosignatures:

   broker_web.apps.utils.templatetags.utility_tags.bytes_to_64utf8
   broker_web.apps.utils.templatetags.utility_tags.jd_to_readable_date
   broker_web.apps.utils.templatetags.utility_tags.urlparams
"""

from base64 import b64decode, b64encode
import fastavro
from io import BytesIO
import json
from urllib.parse import urlencode

from astropy.time import Time
from django import template

register = template.Library()


@register.simple_tag
def urlparams(**kwargs):
    """Format keyword arguments as url parameters.

    Returns:
        A string in the format '?<key>=<value>'
    """
    safe_args = {k: v for k, v in kwargs.items() if v is not None}
    if safe_args:
        return '?{}'.format(urlencode(safe_args))

    return ''


@register.filter
def bytes_to_64utf8(bytes_data):
    """Convert bytes data to UTF8.

    Args:
        bytes_data (bytes): Bytes data

    Returns:
        A string in UTF-8 format
    """
    if bytes_data is not None:
        return b64encode(bytes_data).decode('utf-8')


@register.filter
def base64_to_dict(bytes_data):
    """Convert base64 encoded bytes data to a dict.

    Args:
        bytes_data (base64 bytes): base64 encoded bytes

    Returns:
        A dictionary
    """
    if bytes_data is not None:
        return json.loads(b64decode(bytes_data).decode("utf-8"))


@register.filter
def avro_to_dict(bytes_data):
    """Convert Avro serialized bytes data to a dict.

    Args:
        bytes_data (bytes): Avro serialized bytes

    Returns:
        A dictionary
    """
    if bytes_data is not None:
        with BytesIO(bytes_data) as fin:
            alert_dicts = [r for r in fastavro.reader(fin)]  # list with single dict
        return alert_dicts[0]


@register.filter
def b64avro_to_dict(bytes_data):
    """Convert base64 encoded, Avro serialized bytes data to a dict.

    Args:
        bytes_data (bytes): base64 encoded, Avro serialized bytes

    Returns:
        A dictionary
    """
    if bytes_data is not None:
        with BytesIO(b64decode(bytes_data)) as fin:
            alert_dicts = [r for r in fastavro.reader(fin)]  # list with single dict
        return alert_dicts[0]


@register.filter
def jd_to_readable_date(jd):
    """Convert a julian date to a human readable string.

    Args:
        jd (float): Datetime value in julian format

    Returns:
        String in 'day mon year hour:min' format
    """
    return Time(jd, format='jd').strftime("%d %b %Y - %H:%M:%S")



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""Tests for the ``views`` module."""

from django.test import TestCase
from django.urls import reverse

from broker_web.apps.contact import urls, views
from broker_web.apps.contact.forms import ContactForm


class ContactView(TestCase):
    """Tests for the ``Contact`` view"""

    url_name = f'{urls.app_name}:contact'
    template = 'contact/contact_us.html'

    success_url_name = f'{urls.app_name}:contact-sent'
    success_template = 'contact/contact_sent.html'

    def test_get(self):
        """Test ``get`` method returns correct template"""

        url = reverse(self.url_name)
        response = self.client.post(url)

        self.assertEqual(200, response.status_code)
        self.assertTemplateUsed(response, self.template)

    def test_form_valid_redirect(self):
        """Test ``form_valid`` method redirects to correct success URL"""

        self.assertEqual(
            self.success_template,
            ContactView.success_template,
            'Incorrect success template')

        valid_form = ContactForm(data=dict(
            email='test@email.com',
            subject='Test Name',
            message='Test message.'
        ))

        valid_form.is_valid()  # to set the form's ``cleaned_data`` attribute
        view = views.ContactView()
        response = view.form_valid(valid_form)
        self.assertEqual(302, response.status_code, 'Status code is not 302 redirect')

        success_url = reverse(self.success_url_name)
        self.assertEqual(success_url, response.url)


class SuccessView(TestCase):
    """Tests for the ``success_view`` view"""

    url_name = f'{urls.app_name}:contact-sent'
    template = 'contact/contact_sent.html'

    def test_get(self):
        """Test ``get`` method returns correct template"""

        url = reverse(self.url_name)
        response = self.client.get(url)

        self.assertEqual(200, response.status_code)
        self.assertTemplateUsed(response, self.template)



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""Tests for the ``views`` module."""

from django.http import JsonResponse
from django.test import Client, TestCase
from django.urls import reverse

from broker_web.apps.objects import urls


class ObjectsJsonView(TestCase):
    """Tests for the ``ObjectsJsonView`` view"""

    url_name = f'{urls.app_name}:objects-json'
    app_name = urls.app_name

    def setUp(self):
        self.client = Client()

    def test_get_returns_json(self):
        """Test ``get`` method returns Json Response"""

        url = reverse(self.url_name)
        response = self.client.get(url)

        self.assertEqual(200, response.status_code)
        self.assertIsInstance(response, JsonResponse)


class RecentObjectsView(TestCase):
    """Tests for the ``RecentObjects`` view"""

    url_name = f'{urls.app_name}:recent-objects'
    template = 'objects/recent_objects.html'

    def setUp(self):
        self.client = Client()

    def test_get(self):
        """Test ``get`` method returns correct template"""

        url = reverse(self.url_name)
        response = self.client.get(url)

        self.assertEqual(200, response.status_code)
        self.assertTemplateUsed(response, self.template)

    def test_post(self):  # Todo: Add filtering from form and update tests
        """Test ``post`` method returns correct template"""

        url = reverse(self.url_name)
        response = self.client.post(url)

        self.assertEqual(200, response.status_code)
        self.assertTemplateUsed(response, self.template)


class RecentAlertsJsonView(TestCase):
    """Tests for the ``RecentAlertsJsonView`` view"""

    url_name = f'{urls.app_name}:single-object-json'
    app_name = urls.app_name

    def setUp(self):
        self.client = Client()

    def test_get_returns_json(self):
        """Test ``get`` method returns Json Response"""

        url = reverse(self.url_name, kwargs={'pk': 'dummy_id_name'})
        response = self.client.get(url)

        self.assertEqual(200, response.status_code)
        self.assertIsInstance(response, JsonResponse)


class ObjectSummaryView(TestCase):
    """Tests for the ``ObjectSummary`` view"""

    url_name = f'{urls.app_name}:object-summary'
    template = 'objects/object_summary.html'

    def setUp(self):
        self.client = Client()

    def test_get(self):
        """Test ``get`` method returns correct template and object id"""

        dummy_object_id = '123'
        url = reverse(self.url_name, args=[dummy_object_id])
        response = self.client.get(url)

        self.assertEqual(200, response.status_code)
        self.assertTemplateUsed(response, self.template)

        returned_object_id = response.context['object_id']
        self.assertEqual(returned_object_id, dummy_object_id)



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""Tests for the ``views`` module."""

from django.http import JsonResponse
from django.test import Client, TestCase
from django.urls import reverse

from broker_web.apps.alerts import urls


class TestAlertsJson(TestCase):
    """Tests for the ``AlertsJsonView`` view"""

    url_name = f'{urls.app_name}:alerts-json'

    def setUp(self):
        self.client = Client()

    def test_get(self):
        """Test ``get`` method returns Json Response"""

        url = reverse(self.url_name)
        response = self.client.get(url)

        self.assertEqual(200, response.status_code)
        self.assertIsInstance(response, JsonResponse)


class TestRecentAlerts(TestCase):
    """Tests for the ``RecentAlertsView`` view"""

    url_name = f'{urls.app_name}:recent-alerts'
    template = 'alerts/recent-alerts.html'

    def setUp(self):
        self.client = Client()

    def test_get(self):
        """Test ``get`` method returns correct template"""

        url = reverse(self.url_name)
        response = self.client.get(url)

        self.assertEqual(200, response.status_code)
        self.assertTemplateUsed(response, self.template)

    def test_post(self):
        """Test ``post`` method returns correct template"""

        url = reverse(self.url_name)
        response = self.client.post(url)

        self.assertEqual(200, response.status_code)
        self.assertTemplateUsed(response, self.template)


class TestAlertSummary(TestCase):
    """Tests for the ``AlertSummaryView`` view"""

    url_name = f'{urls.app_name}:alert-summary'
    template = 'alerts/alert_summary.html'
    template_404 = 'alerts/error_404.html'

    def setUp(self):
        self.client = Client()

    def test_app_specific_404_invalid_alert_id(self):
        """Test ``get`` method returns correct template and alert id"""

        dummy_alert_id = 123
        url = reverse(self.url_name, args=[dummy_alert_id])
        response = self.client.get(url)

        self.assertEqual(200, response.status_code)
        self.assertTemplateUsed(response, self.template_404)



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Consumer class to pull or query alerts from Pitt-Google."""

from django.conf import settings
from requests_oauthlib import OAuth2Session

from .utils.templatetags.utility_tags import b64avro_to_dict


PITTGOOGLE_PROJECT_ID = "ardent-cycling-243415"


class PittGoogleConsumer:
    """Consumer class to pull or query alerts from Pitt-Google, and manipulate them."""

    def __init__(self, subscription_name):
        """Open a subscriber client. If the subscription doesn't exist, create it.

        View logs:
            1. https://console.cloud.google.com
            2.

        Authentication creates an `OAuth2Session` object which can be used to fetch
        data, for example: `response = PittGoogleConsumer.oauth.get({url})`
        """
        self._authenticate()

        # logger
        # TODO: add needed params/logic

        # subscription or table resource
        # TODO: add needed params/logic
        self._get_resource()

    def _authenticate(self):
        """Authenticate the user via OAuth 2.0.

        The user will need to visit a URL and authorize `PittGoogleConsumer` to manage
        resources through their Google account.
        """
        # create an OAuth2Session
        client_id = settings.PITTGOOGLE_OAUTH_CLIENT_ID
        client_secret = settings.PITTGOOGLE_OAUTH_CLIENT_SECRET
        redirect_uri = "https://ardent-cycling-243415.appspot.com/"  # TODO: better page
        scopes = [
            "https://www.googleapis.com/auth/logging.write",
        ]
        oauth = OAuth2Session(client_id, redirect_uri=redirect_uri, scope=scopes)

        # instruct the user to authorize
        authorization_url, state = oauth.authorization_url(
            "https://accounts.google.com/o/oauth2/auth",
            access_type="offline",
            prompt="select_account",
        )
        print(
            f"Please visit this URL to authorize PittGoogleConsumer:\n{authorization_url}"
        )
        authorization_response = input(
            "Enter the full URL of the page you are redirected to after authorization:\n"
        )

        # complete the authentication
        token = oauth.fetch_token(
            "https://accounts.google.com/o/oauth2/token",
            authorization_response=authorization_response,
            client_secret=client_secret,
        )
        self.oauth = oauth

    def _get_resource(self):
        """Make sure the resource exists, and we can connect to it."""
        # TODO: add needed logic
        return

    def unpack_messages(
        self, response, lighten_alerts=False, callback=None, **kwargs
    ):
        """Unpack messages in `response`. Run `callback` if present."""
        msgs = response.json()["receivedMessages"]
        alerts = []
        for msg in msgs:
            alert_dict = b64avro_to_dict(msg["message"]["data"])

            if lighten_alerts:
                alert_dict = self._lighten_alert(alert_dict)

            if callback is not None:
                alert_dict = callback(alert_dict, **kwargs)

            if alert_dict is not None:
                alerts.append(alert_dict)

        return alerts

    def _lighten_alert(self, alert_dict):
        keep_fields = {
            "top-level": ["objectId", "candid", ],
            "candidate": ["jd", "ra", "dec", "magpsf", "classtar", ],
        }
        alert_lite = {k: alert_dict[k] for k in keep_fields["top-level"]}
        alert_lite.update(
            {k: alert_dict["candidate"][k] for k in keep_fields["candidate"]}
        )
        return alert_lite

    def _log_and_print(self, msg, severity="INFO"):
        # request = {
        #     'logName': self.log_name,
        #     'resource': {
        #         'type': 'pubsub_subscription',
        #         'labels': {
        #             'project_id': settings.GOOGLE_CLOUD_PROJECT,
        #             'subscription_id': self.subscription_name
        #         },
        #     },
        #     'entries': [{'textPayload': msg, 'severity': severity}],
        # }
        # response = self.oauth.post(self.logging_url, json=json.dumps(request))
        # print(response.content)
        # response.raise_for_status()
        print(msg)



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""The ``managers`` module defines ``Manager`` objects to handle DB changes"""

from django.contrib.auth.base_user import BaseUserManager


class CustomUserManager(BaseUserManager):
    """Custom user model manager"""

    def create_user(self, email, password, **extra_fields):
        """Create and save a User with the given email and password."""

        extra_fields.setdefault('is_active', False)
        user = self.model(email=email, **extra_fields)
        user.set_password(password)
        user.save()
        return user

    def create_superuser(self, email, password, **extra_fields):
        """Create and save a SuperUser with the given email and password."""

        extra_fields.setdefault('is_staff', True)
        extra_fields.setdefault('is_superuser', True)
        extra_fields.setdefault('is_active', True)

        if extra_fields.get('is_staff') is not True:
            raise ValueError('Superuser must have is_staff=True.')

        if extra_fields.get('is_superuser') is not True:
            raise ValueError('Superuser must have is_superuser=True.')

        return self.create_user(email, password, **extra_fields)



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""Tests for the ``urls`` module."""

from django.test import TestCase
from django.urls import resolve, reverse

from broker_web.apps.alerts import views
from broker_web.apps.objects import urls, views


class TestUrlRouting(TestCase):
    """Test URLs are routed to the correct views"""

    app_name = urls.app_name

    def test_objects_json_routing(self):
        """Test 'objects-json' is routed to``ObjectsJsonView``"""

        url = reverse(f'{self.app_name}:objects-json')
        self.assertEqual(views.RecentObjectsJsonView, resolve(url).func.view_class)

    def test_recent_objects_routing(self):
        """Test 'recent-objects' is routed to``RecentObjectsView``"""

        url = reverse(f'{self.app_name}:recent-objects')
        self.assertEqual(views.RecentObjectsView, resolve(url).func.view_class)

    def test_object_summary_routing(self):
        """Test 'object-summary' is routed to``ObjectSummaryView``"""

        dummy_alert_pk = '123'
        url = reverse(f'{self.app_name}:object-summary', args=[dummy_alert_pk])
        self.assertEqual(views.ObjectSummaryView, resolve(url).func.view_class)



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""TOM Toolkit broker to listen to a Pitt-Google Pub/Sub stream via the REST API.

Relies on `ConsumerStreamRest` to manage the connections and work with data.

See especially:

.. autosummary::
   :nosignatures:

   BrokerStreamRest.request_alerts
   BrokerStreamRest.user_filter

"""

from django import forms
from tom_alerts.alerts import GenericQueryForm, GenericAlert, GenericBroker

from .consumer_stream_rest import ConsumerStreamRest
from .utils.templatetags.utility_tags import jd_to_readable_date


class FilterAlertsForm(GenericQueryForm):
    """Basic form for filtering alerts.

    Fields:

        subscription_name (``CharField``)

        classtar_threshold (``FloatField``)

        classtar_gt_lt (``ChoiceField``)

        max_results (``IntegerField``)
    """

    subscription_name = forms.CharField(
        required=True,
        initial='ztf-loop',
        help_text=(
            "The subscription will be created if it doesn't already exist "
            "in the user's project. The ztf-loop stream is recommended for testing. "
            "It is a 'heartbeat' stream with ~1 alert/sec."
        )
    )
    classtar_threshold = forms.FloatField(
        required=False,
        min_value=0,
        max_value=1,
        help_text="Star/Galaxy score threshold",
    )
    classtar_gt_lt = forms.ChoiceField(
        required=True,
        choices=[("lt", "less than"), ("gt", "greater than or equal")],
        initial="lt",
        widget=forms.RadioSelect,
        label="",
    )
    max_results = forms.IntegerField(
        required=False,
        initial=100,
        min_value=1,
    )


class BrokerStreamRest(GenericBroker):
    """Pitt-Google broker class to pull alerts from a stream via the REST API.

    Base class: ``tom_alerts.alerts.GenericBroker``
    """

    name = "Pitt-Google StreamRest"
    form = FilterAlertsForm

    def fetch_alerts(self, parameters):
        """Entry point to pull and filter alerts."""
        clean_params = self._clean_parameters(parameters)

        self.consumer = ConsumerStreamRest(clean_params['subscription_name'])

        alerts, i, max_tries = [], 0, 5  # avoid trying forever
        while (len(alerts) < parameters['max_results']) & (i < max_tries):
            i += 1
            clean_params['max_results'] = parameters['max_results'] - len(alerts)
            alerts += self.request_alerts(clean_params)  # List[dict]

        return iter(alerts)

    def request_alerts(self, parameters):
        """Pull alerts using a POST request with OAuth2, unpack, apply user filter.

        Returns:
            alerts (List[dict])
        """
        response = self.consumer.oauth2.post(
            f"{self.consumer.subscription_url}:pull",
            data={"maxMessages": parameters["max_results"]},
        )
        response.raise_for_status()
        alerts = self.consumer.unpack_and_ack_messages(
            response,
            lighten_alerts=True,
            callback=self.user_filter,
            parameters=parameters,
        )  # List[dict]
        return alerts

    @staticmethod
    def user_filter(alert_dict, parameters):
        """Apply the filter indicated by the form's parameters.

        Used as the `callback` to `BrokerStreamRest.unpack_and_ack_messages`.

        Args:
            `alert_dict`: Single alert, ZTF packet data as a dictionary.
                          The schema depends on the value of `lighten_alerts` passed to
                          `BrokerStreamRest.unpack_and_ack_messages`.
                          If `lighten_alerts=False` it is the original ZTF alert schema
                          (https://zwickytransientfacility.github.io/ztf-avro-alert/schema.html).
                          If `lighten_alerts=True` the dict is flattened and extra
                          fields are dropped.

            `parameters`: parameters submitted by the user through the form.

        Returns:
            `alert_dict` if it passes the filter, else `None`
        """
        if parameters["classtar_threshold"] is None:
            # no filter requested. all alerts pass
            return alert_dict

        # run the filter
        lt_threshold = alert_dict["classtar"] < parameters["classtar_threshold"]
        if ((parameters["classtar_gt_lt"] == "lt") & lt_threshold) or (
            (parameters["classtar_gt_lt"] == "gt") & ~lt_threshold
        ):
            return alert_dict
        else:
            return None

    def _clean_parameters(self, parameters):
        clean_params = dict(parameters)
        return clean_params

    def to_generic_alert(self, alert):
        """Map the Pitt-Google alert to a TOM `GenericAlert`."""
        return GenericAlert(
            timestamp=jd_to_readable_date(alert["jd"]),
            url=self.consumer.subscription_url,
            id=alert["candid"],
            name=alert["objectId"],
            ra=alert["ra"],
            dec=alert["dec"],
            mag=alert["magpsf"],
            score=alert["classtar"],
        )



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""Tests for the ``models`` and ``managers`` modules"""

from django.contrib.auth import get_user_model
from django.test import TestCase


class CustomUserManager(TestCase):
    """Test user creation"""

    def test_create_user(self):
        """Test creation of regular user has correct defaults"""

        user_model = get_user_model()

        test_email = 'test@user.com'
        new_user = user_model.objects.create_user(email=test_email, password='test')

        self.assertEqual(new_user.email, test_email, 'Email does not equal passed value.')
        self.assertFalse(new_user.is_active, 'User is set to active')
        self.assertFalse(new_user.is_staff, 'User is staff')
        self.assertFalse(new_user.is_superuser, 'User is superuser')

        with self.assertRaises(TypeError, msg='No error raised for missing email'):
            user_model.objects.create_user()

        with self.assertRaises(TypeError, msg='No error raised for blank email'):
            user_model.objects.create_user(email='')

    def test_create_superuser(self):
        """Test creation of a superuser has correct defaults"""

        user_model = get_user_model()

        test_email = 'test@user.com'
        new_admin_user = user_model.objects.create_superuser(test_email, 'foo')

        self.assertEqual(new_admin_user.email, test_email, 'Email does not equal passed value.')
        self.assertTrue(new_admin_user.is_active, 'User is not active')
        self.assertTrue(new_admin_user.is_staff, 'User is not staff')
        self.assertTrue(new_admin_user.is_superuser, 'User is not superuser')

        with self.assertRaises(TypeError, msg='No error raised for missing email'):
            user_model.objects.create_superuser()

        with self.assertRaises(TypeError, msg='No error raised for blank email'):
            user_model.objects.create_superuser(email='')

        with self.assertRaises(ValueError, msg='No error raised for is_superuser=False'):
            user_model.objects.create_superuser(
                email='super@user.com', password='foo', is_superuser=False)



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""Adds commands to the default Django management script ``manage.py``

For more information see:
  https://docs.djangoproject.com/en/3.0/howto/custom-management-commands/
"""



"""A setuptools based setup module.

See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

long_description = (here / 'README.md').read_text(encoding='utf-8')

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name='tom_pittgoogle',  # Required
    version='0.1.8',  # Required
    description='Pitt-Google broker module for the TOM Toolkit',  # Optional
    long_description=long_description,  # Optional
    long_description_content_type='text/markdown',  # Optional (see note above)
    # url='https://github.com/pypa/sampleproject',  # Optional

    author='Troy Raen',  # Optional
    author_email='troy.raen@pitt.edu',  # Optional
    classifiers=[  # Optional
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Astronomy',
        'Topic :: Scientific/Engineering :: Physics',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
    ],
    keywords='astronomy, alert broker',
    packages=find_packages(),  # Required
    python_requires='>=3.6, <4',
    install_requires=requirements,  # Optional
)



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Consumer class to manage BigQuery connections via Python client, and work with data.

BigQuery Python Client docs: https://googleapis.dev/python/bigquery/latest/index.html

Used by `BrokerDatabasePython`, but can be called independently.

Basic workflow:

.. code:: python

    consumer = ConsumerDatabasePython(table_name)

    sql_stmnt, job_config = consumer.create_sql_stmnt(parameters)
    query_job = consumer.client.query(sql_stmnt, job_config=job_config)

    alerts = consumer.unpack_query(query_job)  # List[dict]

See especially:

.. autosummary::
   :nosignatures:

   ConsumerDatabasePython.authenticate
   ConsumerDatabasePython.create_sql_stmnt
   ConsumerDatabasePython.unpack_query

"""

from django.conf import settings
from google.api_core.exceptions import NotFound
from google.cloud import bigquery
from google.cloud import logging as gc_logging
from google_auth_oauthlib.helpers import credentials_from_session
from requests_oauthlib import OAuth2Session


PITTGOOGLE_PROJECT_ID = "ardent-cycling-243415"


class ConsumerDatabasePython:
    """Consumer class to query alerts from BigQuery, and manipulate them.

    Initialization does the following:

        - Authenticate the user via OAuth 2.0.

        - Create a `google.cloud.bigquery.Client` object for the user/broker
          to query database with.

        - Check that the table exists and we can connect.

    To view logs, visit: https://console.cloud.google.com/logs

        - Make sure you are logged in, and your project is selected in the dropdown
          at the top.

        - Click the "Log name" dropdown and select the table name you instantiate this
          consumer with.

    TODO: Give the user a standard logger.
    """

    def __init__(self, table_name):
        """Authenticate user; create a client, the table path and check connection."""
        user_project = settings.GOOGLE_CLOUD_PROJECT

        self.authenticate()
        self.credentials = credentials_from_session(self.oauth2)
        self.client = bigquery.Client(
            project=user_project, credentials=self.credentials
        )

        # logger
        log_client = gc_logging.Client(
            project=user_project, credentials=self.credentials
        )
        self.logger = log_client.logger(table_name)

        # table
        self.table_name = table_name
        self.table_path = f"{PITTGOOGLE_PROJECT_ID}.{table_name}"
        self._get_table()

        # for the TOM `GenericAlert`. this won't be very helpful without instructions.
        self.query_url = f"https://bigquery.googleapis.com/bigquery/v2/projects/{user_project}/queries"

    def authenticate(self):
        """Guide user through authentication; create `OAuth2Session` for credentials.

        The user will need to visit a URL, authenticate themselves, and authorize
        `PittGoogleConsumer` to make API calls on their behalf.

        The user must have a Google account that is authorized make API calls
        through the project defined by the `GOOGLE_CLOUD_PROJECT` variable in the
        Django `settings.py` file. Any project can be used, as long as the user has
        access.

        Additional requirement because this is still in dev: The OAuth is restricted
        to users registered with Pitt-Google, so contact us.

        TODO: Integrate this with Django. For now, the user interacts via command line.
        """
        # create an OAuth2Session
        client_id = settings.PITTGOOGLE_OAUTH_CLIENT_ID
        client_secret = settings.PITTGOOGLE_OAUTH_CLIENT_SECRET
        authorization_base_url = "https://accounts.google.com/o/oauth2/auth"
        redirect_uri = "https://ardent-cycling-243415.appspot.com/"  # TODO: better page
        scopes = [
            "https://www.googleapis.com/auth/logging.write",
            "https://www.googleapis.com/auth/bigquery",
        ]
        oauth2 = OAuth2Session(client_id, redirect_uri=redirect_uri, scope=scopes)

        # instruct the user to authorize
        authorization_url, state = oauth2.authorization_url(
            authorization_base_url,
            access_type="offline",
            # access_type="online",
            # prompt="select_account",
        )
        print((
            "Please visit this URL to authenticate yourself and authorize "
            "PittGoogleConsumer to make API calls on your behalf:"
            f"\n\n{authorization_url}\n"
        ))
        authorization_response = input(
            "After authorization, you should be directed to the Pitt-Google Alert "
            "Broker home page. Enter the full URL of that page (it should start with "
            "https://ardent-cycling-243415.appspot.com/):\n"
        )

        # complete the authentication
        _ = oauth2.fetch_token(
            "https://accounts.google.com/o/oauth2/token",
            authorization_response=authorization_response,
            client_secret=client_secret,
        )
        self.oauth2 = oauth2

    def _get_table(self):
        """Make sure the resource exists, and we can connect to it."""
        try:
            _ = self.client.get_table(self.table_path)
        except NotFound as e:
            self._log_and_print(
                f"Table not found. Can't connect to {self.table_path}", severity="DEBUG"
            )
            raise e
        else:
            self._log_and_print(f"Successfully connected to {self.table_path}")

    def create_sql_stmnt(self, parameters):
        """Create the SQL statement and a job config with the user's `parameters`."""
        query_parameters = []

        select = """
            SELECT
                publisher,
                objectId,
                candid,
                CASE candidate.fid
                    WHEN 1 THEN 'g' WHEN 2 THEN 'R' WHEN 3 THEN 'i' END as filter,
                candidate.magpsf as mag,
                candidate.jd as jd,
                candidate.ra as ra,
                candidate.dec as dec,
                candidate.classtar as classtar
        """

        frm = f"FROM `{self.table_path}`"

        where = ""
        if parameters["objectId"]:
            where = "WHERE objectId=@objectId"
            query_parameters.append(
                bigquery.ScalarQueryParameter(
                    "objectId", "STRING", parameters["objectId"]
                )
            )
        elif parameters["candid"]:
            where = "WHERE candid=@candid"
            query_parameters.append(
                bigquery.ScalarQueryParameter("candid", "INT64", parameters["candid"])
            )

        orderby = "ORDER BY jd DESC"

        limit = "LIMIT @max_results"
        query_parameters.append(
            bigquery.ScalarQueryParameter(
                "max_results", "INT64", parameters["max_results"]
            )
        )

        # construct the SQL statement and job config
        sql_stmnt = " ".join([select, frm, where, orderby, limit])
        job_config = bigquery.QueryJobConfig(query_parameters=query_parameters)

        return sql_stmnt, job_config

    def unpack_query(
        self, query_job, callback=None, **kwargs
    ):
        """Unpack alerts from `query_job`; run `callback` if present.

        A basic filter is implemented directly in the SQL statement produced by
        `create_sql_stmnt`. More complex filters could be implemented here via a
        `callback` function.
        """
        alerts = []
        for row in query_job.result():
            alert_dict = dict(row)

            if callback is not None:
                alert_dict = callback(alert_dict, **kwargs)

            if alert_dict is not None:
                alerts.append(alert_dict)

        return alerts

    def _log_and_print(self, msg, severity="INFO"):
        self.logger.log_text(msg, severity=severity)
        print(msg)



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Consumer class to manage Pub/Sub connections via REST, and work with message data.

Pub/Sub REST API docs: https://cloud.google.com/pubsub/docs/reference/rest

Used by `BrokerStreamRest`, but can be called independently.

Basic workflow:

.. code:: python

    consumer = ConsumerStreamRest(subscription_name)

    response = consumer.oauth2.post(
        f"{consumer.subscription_url}:pull", data={"maxMessages": max_messages},
    )

    alerts = consumer.unpack_and_ack_messages(
        response, lighten_alerts=True, callback=user_filter,
    )  # List[dict]

See especially:

.. autosummary::
   :nosignatures:

   ConsumerStreamRest.authenticate
   ConsumerStreamRest.touch_subscription
   ConsumerStreamRest.unpack_and_ack_messages

"""

from django.conf import settings
import json
from requests_oauthlib import OAuth2Session

from .utils.templatetags.utility_tags import b64avro_to_dict


PITTGOOGLE_PROJECT_ID = "ardent-cycling-243415"


class ConsumerStreamRest:
    """Consumer class to manage Pub/Sub connections and work with messages.

    Initialization does the following:

        - Authenticate the user. Create an `OAuth2Session` object for the user/broker
          to make HTTP requests with.

        - Make sure the subscription exists and we can connect. Create it, if needed.
    """

    def __init__(self, subscription_name):
        """Create an `OAuth2Session`. Set `subscription_url` and check connection."""
        self.authenticate()

        # logger
        # TODO: debug the logger
        # the user probably prefers a more standard logger anyway
        # self.logging_url = "https://logging.googleapis.com/v2/entries:write"
        # self.log_name = (
        #     f"projects/{settings.GOOGLE_CLOUD_PROJECT}/logs/{subscription_name}"
        # )

        # subscription
        self.subscription_name = subscription_name
        self.subscription_path = (
            f"projects/{settings.GOOGLE_CLOUD_PROJECT}/subscriptions/{subscription_name}"
        )
        self.subscription_url = (
            f"https://pubsub.googleapis.com/v1/{self.subscription_path}"
        )
        self.topic_path = ""  # for user info only. set in touch_subscription()
        self.touch_subscription()

    def authenticate(self):
        """Guide user through authentication; create `OAuth2Session` for HTTP requests.

        The user will need to visit a URL, authenticate themselves, and authorize
        `PittGoogleConsumer` to make API calls on their behalf.

        The user must have a Google account that is authorized make API calls
        through the project defined by the `GOOGLE_CLOUD_PROJECT` variable in the
        Django `settings.py` file. Any project can be used, as long as the user has
        access.

        Additional requirement because this is still in dev: The OAuth is restricted
        to users registered with Pitt-Google, so contact us.

        TODO: Integrate this with Django. For now, the user interacts via command line.
        """
        # create an OAuth2Session
        client_id = settings.PITTGOOGLE_OAUTH_CLIENT_ID
        client_secret = settings.PITTGOOGLE_OAUTH_CLIENT_SECRET
        redirect_uri = "https://ardent-cycling-243415.appspot.com/"  # TODO: better page
        scopes = [
            "https://www.googleapis.com/auth/logging.write",
            "https://www.googleapis.com/auth/pubsub",
        ]
        oauth2 = OAuth2Session(client_id, redirect_uri=redirect_uri, scope=scopes)

        # instruct the user to authorize
        authorization_url, state = oauth2.authorization_url(
            "https://accounts.google.com/o/oauth2/auth",
            access_type="offline",
            # access_type="online",
            # prompt="select_account",
        )
        print((
            "Please visit this URL to authenticate yourself and authorize "
            "PittGoogleConsumer to make API calls on your behalf:"
            f"\n\n{authorization_url}\n"
        ))
        authorization_response = input(
            "After authorization, you should be directed to the Pitt-Google Alert "
            "Broker home page. Enter the full URL of that page (it should start with "
            "https://ardent-cycling-243415.appspot.com/):\n"
        )

        # complete the authentication
        _ = oauth2.fetch_token(
            "https://accounts.google.com/o/oauth2/token",
            authorization_response=authorization_response,
            client_secret=client_secret,
        )
        self.oauth2 = oauth2

    def touch_subscription(self):
        """Make sure the subscription exists and we can connect.

        If the subscription doesn't exist, try to create one (in the user's project)
        that is attached to a topic of the same name in the Pitt-Google project.

        Note that messages published before the subscription is created are not
        available.
        """
        # check if subscription exists
        response = self.oauth2.get(self.subscription_url)

        if response.status_code == 200:
            # subscription exists. tell the user which topic it's connected to.
            self.topic_path = json.loads(response.content)["topic"]
            print(f"Subscription exists: {self.subscription_path}")
            print(f"Connected to topic: {self.topic_path}")

        elif response.status_code == 404:
            # subscription doesn't exist. try to create it.
            self._create_subscription()

        else:
            print(response.content)
            response.raise_for_status()

    def _create_subscription(self):
        """Try to create the subscription."""
        topic_path = f"projects/{PITTGOOGLE_PROJECT_ID}/topics/{self.subscription_name}"
        request = {"topic": topic_path}
        put_response = self.oauth2.put(f"{self.subscription_url}", data=request)

        if put_response.status_code == 200:
            # subscription created successfully
            self.topic_path = topic_path
            self._log_and_print(
                (
                    f"Created subscription: {self.subscription_path}\n"
                    f"Connected to topic: {self.topic_path}"
                )
            )

        elif put_response.status_code == 404:
            raise ValueError(
                (
                    f"A subscription named {self.subscription_name} does not exist"
                    "in the Google Cloud Platform project "
                    f"{settings.GOOGLE_CLOUD_PROJECT}, "
                    "and one cannot be create because Pitt-Google does not "
                    "publish a public topic with the same name."
                )
            )

        else:
            # if the subscription name is invalid, content has helpful info
            print(put_response.content)
            put_response.raise_for_status()

    def unpack_and_ack_messages(
        self, response, lighten_alerts=False, callback=None, **kwargs
    ):
        """Unpack and acknowledge messages in `response`; run `callback` if present.

        If `lighten_alerts` is True, drop extra fields and flatten the alert dict.

        `callback` is assumed to be a filter. It should accept an alert dict
        and return the dict if the alert passes the filter, else return None.
        """
        # unpack and run the callback
        try:
            msgs = response.json()["receivedMessages"]
        except KeyError:
            msg = (
                "No messages received. If you recently created the subscription, "
                "it's possible that no messages have been published to the "
                "topic since the subscription's creation."
                "Try connecting to the heartbeat stream ztf-loop."
            )
            raise ValueError(msg)

        alerts, ack_ids = [], []
        for msg in msgs:
            alert_dict = b64avro_to_dict(msg["message"]["data"])

            if lighten_alerts:
                alert_dict = self._lighten_alert(alert_dict)

            if callback is not None:
                alert_dict = callback(alert_dict, **kwargs)

            if alert_dict is not None:
                alerts.append(alert_dict)
            ack_ids.append(msg["ackId"])

        # acknowledge messages so they leave the subscription
        self._ack_messages(ack_ids)

        return alerts

    def _lighten_alert(self, alert_dict):
        """Return the alert as a flat dict, keeping only the fields in `keep_fields`."""
        keep_fields = {
            "top-level": ["objectId", "candid", ],
            "candidate": ["jd", "ra", "dec", "magpsf", "classtar", ],
        }
        alert_lite = {k: alert_dict[k] for k in keep_fields["top-level"]}
        alert_lite.update(
            {k: alert_dict["candidate"][k] for k in keep_fields["candidate"]}
        )
        return alert_lite

    def _ack_messages(self, ack_ids):
        """Send message acknowledgements to Pub/Sub."""
        response = self.oauth2.post(
            f"{self.subscription_url}:acknowledge", data={"ackIds": ack_ids},
        )
        response.raise_for_status()

    def delete_subscription(self):
        """Delete the subscription.

        This is provided for the user's convenience, but it is not necessary and is not
        automatically called.

            - Storage of unacknowledged Pub/Sub messages does not result in fees.

            - Unused subscriptions automatically expire; default is 31 days.
        """
        response = self.oauth2.delete(self.subscription_url)
        if response.status_code == 200:
            self._log_and_print(f"Deleted subscription: {self.subscription_path}")
        elif response.status_code == 404:
            print(
                f"Nothing to delete, subscription does not exist: {self.subscription_path}"
            )
        else:
            response.raise_for_status()

    def _log_and_print(self, msg, severity="INFO"):
        # TODO: fix this
        # request = {
        #     'logName': self.log_name,
        #     'resource': {
        #         'type': 'pubsub_subscription',
        #         'labels': {
        #             'project_id': settings.GOOGLE_CLOUD_PROJECT,
        #             'subscription_id': self.subscription_name
        #         },
        #     },
        #     'entries': [{'textPayload': msg, 'severity': severity}],
        # }
        # response = self.oauth2.post(self.logging_url, json=json.dumps(request))
        # print(response.content)
        # response.raise_for_status()
        print(msg)



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""The ``utility_tags`` module defines functions for casting between
different data types and formats. These functions are automatically
registered as template tags for use in the Django template engine.

.. autosummary::
   :nosignatures:

   broker_web.apps.utils.templatetags.utility_tags.bytes_to_64utf8
   broker_web.apps.utils.templatetags.utility_tags.jd_to_readable_date
   broker_web.apps.utils.templatetags.utility_tags.urlparams
"""

from base64 import b64encode
from urllib.parse import urlencode

from astropy.time import Time
from django import template

register = template.Library()


@register.simple_tag
def urlparams(**kwargs):
    """Format keyword arguments as url parameters

    Returns:
        A string in the format '?<key>=<value>'
    """

    safe_args = {k: v for k, v in kwargs.items() if v is not None}
    if safe_args:
        return '?{}'.format(urlencode(safe_args))

    return ''


@register.filter
def bytes_to_64utf8(bytes_data):
    """Convert bytes data to UTF8

    Args:
        bytes_data (bytes): Bytes data

    Returns:
        A string in UTF-8 format
    """

    if bytes_data is not None:
        return b64encode(bytes_data).decode('utf-8')


@register.filter
def jd_to_readable_date(jd):
    """Convert a julian date to a human readable string

    Args:
        jd (float): Datetime value in julian format

    Returns:
        String in 'day mon year hour:min' format
    """

    return Time(jd, format='jd').strftime("%d %b %Y - %H:%M:%S")



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""Tests for the ``views`` module."""

from django.contrib.auth import get_user_model
from django.test import Client, TestCase
from django.urls import reverse

from broker_web.apps.subscriptions import urls

User = get_user_model()


class GenericTests:
    """Tests that are applicable to multiple views"""

    def test_unauthenticated_get(self):
        """Test ``get`` method returns correct template with 200 status code"""

        # Define the url for the view, and the login url that the user should
        # be redirected to for being unauthenticated.
        view_url = reverse(self.url_name)
        login_url = reverse('login') + '?next=' + view_url

        response = Client().get(view_url)
        self.assertEqual(302, response.status_code)
        self.assertEqual(login_url, response.url)

    def test_authenticated_get(self):
        """Test ``get`` method returns correct template with 200 status code"""

        test_user = User.objects.get_or_create(email='user@email.com')[0]
        client = Client()
        client.force_login(test_user)

        url = reverse(self.url_name)
        response = client.get(url)
        self.assertEqual(200, response.status_code)
        self.assertTemplateUsed(response, self.template)


class SubscriptionsView(TestCase, GenericTests):
    """Tests for the ``Subscriptions`` view"""

    url_name = f'{urls.app_name}:subscriptions'
    template = 'subscriptions/subscriptions.html'


class ProfileView(TestCase, GenericTests):
    """Tests for the ``Profile`` view"""

    url_name = f'{urls.app_name}:profile'
    template = 'subscriptions/my_profile.html'



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""TOM Toolkit broker to listen to a Pitt-Google Pub/Sub stream via the Python client.

Relies on `ConsumerStreamPython` to manage the connections and work with data.

See especially:

.. autosummary::
   :nosignatures:

   BrokerStreamPython.fetch_alerts
   BrokerStreamPython.user_filter

"""
from django import forms
from tom_alerts.alerts import GenericQueryForm, GenericAlert, GenericBroker
from tom_targets.models import Target

from .consumer_stream_python import ConsumerStreamPython
from .utils.templatetags.utility_tags import jd_to_readable_date


class FilterAlertsForm(GenericQueryForm):
    """Basic form for filtering alerts.

    Fields:

        subscription_name (``CharField``)

        classtar_threshold (``FloatField``)

        classtar_gt_lt (``ChoiceField``)

        max_results (``IntegerField``)

        timeout (``IntegerField``)

        max_backlog (``IntegerField``)
    """

    subscription_name = forms.CharField(
        required=True,
        initial='ztf-loop',
        help_text=(
            "The subscription will be created if it doesn't already exist "
            "in the user's project. The ztf-loop stream is recommended for testing. "
            "It is a 'heartbeat' stream with ~1 alert/sec."
        )
    )
    classtar_threshold = forms.FloatField(
        required=False,
        min_value=0,
        max_value=1,
        help_text="Star/Galaxy score threshold",
    )
    classtar_gt_lt = forms.ChoiceField(
        required=True,
        choices=[("lt", "less than"), ("gt", "greater than or equal")],
        initial="lt",
        widget=forms.RadioSelect,
        label="",
    )
    max_results = forms.IntegerField(
        required=False,
        initial=100,
        min_value=1,
        help_text=(
            "Maximum number of alerts to pull and process before stopping the "
            "streaming pull. Recommended for testing only."
        )
    )
    timeout = forms.IntegerField(
        required=False,
        initial=30,
        min_value=1,
        help_text=(
            "Maximum amount of time in seconds to wait for a new alert before stopping "
            "the streaming pull. Recommended for testing to avoid waiting for "
            "max number of results forever. "
            "Recommended for production, possibly in combination "
            "with a time of day-based stopping condition."
        )
    )
    max_backlog = forms.IntegerField(
        required=False,
        initial=1000,
        min_value=1,
        help_text=(
            "Maximum number of pulled but unprocessed alerts before pausing the "
            "streaming pull. Google's default is 1000, which is often fine. "
            "However, you may want to reduce this if you set 'Max results' to a "
            "small number and you want to avoid pulling down a bunch of extra alerts. "
            "Note that alerts do not 'leave' the subscription until they are "
            "successfully processed by the callback; this setting does not affect that."
        )
    )


class BrokerStreamPython(GenericBroker):
    """Pitt-Google broker interface to pull alerts from Pub/Sub via the Python client.

    Base class: ``tom_alerts.alerts.GenericBroker``
    """

    name = "Pitt-Google StreamPython"
    form = FilterAlertsForm

    def fetch_alerts(self, parameters):
        """Entry point to pull and filter alerts.

        Pull alerts using a Python client, unpack, apply user filter.

        This demo assumes that the real use-case is to save alerts to a database
        rather than view them through a TOM site.
        Therefore, the `Consumer` currently saves the alerts in real time,
        and then simply returns a list of alerts after all messages are processed.
        That list is then coerced into an iterator here.
        If the user really cares about the iterator,
        `ConsumerStreamPython.stream_alerts` can be tweaked to yield the alerts in
        real time.
        """
        clean_params = self._clean_parameters(parameters)

        self.consumer = ConsumerStreamPython(clean_params['subscription_name'])

        alert_dicts_list = self.consumer.stream_alerts(
            user_filter=self.user_filter,
            **clean_params,
        )

        return iter(alert_dicts_list)

    def _clean_parameters(self, parameters):
        clean_params = dict(parameters)

        clean_params.setdefault("classtar_threshold", None)

        # there must be at least one stopping condition
        if (clean_params['max_results'] is None) & (clean_params['timeout'] is None):
            raise ValueError((
                "You must set at least one stopping condition. "
                "max_results and timeout cannot both be None."
            ))

        if clean_params['max_backlog'] is None:
            clean_params['max_backlog'] = 1000  # keep the google default of 1000

        return clean_params

    @staticmethod
    def user_filter(alert_dict, **parameters):
        """Apply the filter indicated by the form's parameters.

        Used as the `callback` to `BrokerStreamPython.unpack_and_ack_messages`.

        Args:
            `alert_dict`: Single alert, ZTF packet data as a dictionary.
                          The schema depends on the value of `lighten_alerts` passed to
                          `BrokerStreamPython.unpack_and_ack_messages`.
                          If `lighten_alerts=False` it is the original ZTF alert schema
                          (https://zwickytransientfacility.github.io/ztf-avro-alert/schema.html).
                          If `lighten_alerts=True` the dict is flattened and extra
                          fields are dropped.

            `parameters`: parameters submitted by the user through the form.

        Returns:
            `alert_dict` if it passes the filter, else `None`
        """
        if parameters["classtar_threshold"] is None:
            # no filter requested. all alerts pass
            return alert_dict

        # run the filter
        lt_threshold = alert_dict["classtar"] < parameters["classtar_threshold"]
        if ((parameters["classtar_gt_lt"] == "lt") & lt_threshold) or (
            (parameters["classtar_gt_lt"] == "gt") & ~lt_threshold
        ):
            return alert_dict
        else:
            return None

    def to_generic_alert(self, alert_dict):
        """Map the Pitt-Google alert to a TOM `GenericAlert`."""
        return GenericAlert(
            timestamp=jd_to_readable_date(alert_dict["jd"]),
            # url=self.consumer.pull_url,
            # this is not a valid url
            url="https://pubsub.googleapis.com/v1/{subscription_path}",
            id=alert_dict["candid"],
            name=alert_dict["objectId"],
            ra=alert_dict["ra"],
            dec=alert_dict["dec"],
            mag=alert_dict["magpsf"],
            score=alert_dict["classtar"],
        )

    def to_target(self, alert_dict):
        """Map the Pitt-Google alert to a TOM `Target`."""
        return Target(
            # identifier=alert_dict['candid'],
            name=alert_dict['objectId'],
            type='SIDEREAL',
            # designation='MY ALERT',
            ra=alert_dict['ra'],
            dec=alert_dict['dec'],
            # epoch=alert_dict['jd'],
        )

    # def get_or_create_target(self, alert_dict):
    #     target, created = Target.objects.get_or_create(
    #         name=alert_dict['objectId'],
    #         type='SIDEREAL',
    #         ra=alert_dict['ra'],
    #         dec=alert_dict['dec'],
    #     )



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""Tests for the ``templatetags.utility_tags`` module."""

from django.test import TestCase

from broker_web.apps.utils.templatetags.utility_tags import jd_to_readable_date
from broker_web.apps.utils.templatetags.utility_tags import urlparams


class JDToReadableDate(TestCase):
    """Tests for the ``jd_to_readable_date`` tag"""

    def test_date_conversion(self):
        """Test the date is returned as a string for a pre-chosen JD"""

        jd = 2459145.19
        readable_string = '22 Oct 2020 - 16:33:36'
        self.assertEqual(jd_to_readable_date(jd), readable_string)


class URLParams(TestCase):
    """Tests for the ``urlparams`` tag"""

    def test_empty_args(self):
        """Test an empty string is returned for empty kwargs"""

        self.assertEqual(urlparams(), '')

    def test_multiple_kwargs_combined_with_ampersand(self):
        """Test multiple arguments are combined with & symbol"""

        kwargs = {'arg1': 'val1', 'arg2': 'val2'}
        expected_return = '?arg1=val1&arg2=val2'
        self.assertEqual(urlparams(**kwargs), expected_return)



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""The ``views`` module defines ``View`` objects for converting web requests
into rendered responses.

.. autosummary::
   :nosignatures:

   broker_web.apps.contact.views.ContactView
"""

from django.conf import settings
from django.core.mail import BadHeaderError, send_mail
from django.http import HttpResponse
from django.urls import reverse_lazy
from django.views.generic import FormView, TemplateView

from .forms import ContactForm

SuccessView = TemplateView.as_view(template_name='contact/contact_sent.html')


class ContactView(FormView):
    """View for submitting an email to the website maintainers"""

    template_name = "contact/contact_us.html"
    form_class = ContactForm
    # Todo: Include app name in reverse lookup in case app is namespaced differently
    success_url = reverse_lazy('contact:contact-sent')

    def form_valid(self, form):
        """Send contents of email form and redirect to success url

        Called after form is validated

        Args:
            form (django.forms.Form): User creation form
        """

        subject = form.cleaned_data['subject']
        email = form.cleaned_data['email']
        message = form.cleaned_data['message']

        try:
            send_mail(subject, message, email, settings.CONTACT_EMAILS)

        except BadHeaderError:
            return HttpResponse('Invalid header found.')

        return super().form_valid(form)



# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""Tests for the ``urls`` module."""

from django.test import TestCase
from django.urls import resolve, reverse

from broker_web.apps.getting_started import urls, views


class TestUrlRouting(TestCase):
    """Test URLs are routed to the correct views"""

    app_name = urls.app_name

    def test_introduction_routing(self):
        """Test 'introduction' is routed to ``Introduction``"""

        url = reverse(f'{self.app_name}:introduction')
        self.assertEqual(views.Introduction, resolve(url).func)

    def test_data_products_routing(self):
        """Test 'data-products' is routed to ``DataProducts``"""

        url = reverse(f'{self.app_name}:data-products')
        self.assertEqual(views.DataProducts, resolve(url).func)

    def test_technical_resources_routing(self):
        """Test 'technical-resources' is routed to ``TechnicalResources``"""

        url = reverse(f'{self.app_name}:technical-resources')
        self.assertEqual(views.TechnicalResources, resolve(url).func)

    def test_data_access_routing(self):
        """Test 'data-access' is routed to ``DataAccess``"""

        url = reverse(f'{self.app_name}:data-access')
        self.assertEqual(views.DataAccess, resolve(url).func)

    def test_broker_design_routing(self):
        """Test 'broker-design' is routed to ``BrokerDesign``"""

        url = reverse(f'{self.app_name}:broker-design')
        self.assertEqual(views.BrokerDesign, resolve(url).func)




# !/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""The ``views`` module defines ``View`` objects for converting web requests
into rendered responses.

.. autosummary::
   :nosignatures:

   broker_web.apps.subscriptions.views.ProfileView
   broker_web.apps.subscriptions.views.SubscriptionsView
"""

from django.contrib.auth.mixins import LoginRequiredMixin
from django.shortcuts import render
from django.views.generic import View


class SubscriptionsView(LoginRequiredMixin, View):
    """View that handles new user subscriptions"""

    def get(self, request, *args, **kwargs):
        return render(request, 'subscriptions/subscriptions.html')


class ProfileView(LoginRequiredMixin, View):
    """View that handles user profiles"""

    def get(self, request, *args, **kwargs):
        """Handle an incoming HTTP request

        Args:
            request (HttpRequest): Incoming HTTP request

        Returns:
            Outgoing HTTPResponse
        """

        # Todo get pubsub messages
        timestamps = [123, 456]
        messages = ['a', 'b']
        context = {
            'pbsub_zip': zip(timestamps, messages)
        }

        return render(request, 'subscriptions/my_profile.html', context)



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Pitt-Google broker module for TOM Toolkit."""

from django import forms
from tom_alerts.alerts import GenericQueryForm, GenericAlert, GenericBroker

from consumer import PittGoogleConsumer


CONSUMER = PittGoogleConsumer()


class FilterAlertsForm(GenericQueryForm):
    """Form for filtering alerts.

    Fields:
        max_results (``IntegerField``)
    """

    max_results = forms.IntegerField(
        required=True, initial=10, min_value=1, max_value=1000
    )


class PittGoogleBroker(GenericBroker):
    """Pitt-Google broker interface to fetch alerts."""

    name = "Pitt-Google"
    form = FilterAlertsForm

    @classmethod
    def fetch_alerts(self, parameters):
        """Pull or query alerts, unpack, apply the user filter, return an iterator."""
        clean_params = self._clean_parameters(parameters)

        alerts = []
        while len(alerts) < parameters['max_results']:
            alerts += self._request_alerts(clean_params)  # List[dict]

        return iter(alerts)

    def _request_alerts(self, parameters):
        """Pull or query alerts, unpack, apply the user filter."""
        response = CONSUMER.oauth.post()
        response.raise_for_status()
        alerts = CONSUMER.unpack_messages(
            response,
            lighten_alerts=True,
            callback=self._user_filter,
            parameters=parameters,
        )
        return alerts

    @staticmethod
    def _user_filter(alert, parameters):
        """Apply the filter indicated by the form's parameters.

        Args:
            `alert_dict`: Single alert, ZTF packet data.
            `parameters`: parameters submitted by the user through the form.
        """
        alert_passes_filter = True
        if alert_passes_filter:
            return alert
        else:
            return None

    def _clean_parameters(parameters):
        clean_params = dict(parameters)
        return clean_params

    @classmethod
    def to_generic_alert(self, alert):
        """Map the Pitt-Google alert to a TOM `GenericAlert`."""
        return GenericAlert(
            timestamp=alert["jd"],
            url="https://{service}.googleapis.com/{resource_path}",
            id=alert["candid"],
            name=alert["objectId"],
            ra=alert["ra"],
            dec=alert["dec"],
            mag=alert["magpsf"],
            score=alert["classtar"],
        )
